# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

from lattpy.atom import Atom


def test_atom_uniqueness():
    atom1 = Atom("A")
    atom2 = Atom("A")
    assert atom1 == atom2
    assert atom1.__hash__() == atom2.__hash__()

    atom1 = Atom("A")
    atom2 = Atom("B")
    assert atom1 != atom2
    assert atom1.__hash__() != atom2.__hash__()


def test_atom_params():
    atom = Atom("A", color="r", energy=1)

    assert atom.color == "r"
    assert atom["color"] == "r"

    assert atom.energy == 1
    assert atom["energy"] == 1

    atom["spin"] = 1
    assert atom.spin == 1
    assert atom["spin"] == 1

    atom.spin = 1
    assert atom.spin == 1
    assert atom["spin"] == 1

    assert atom.get("spin") == 1

    del atom["spin"]

    assert atom.get("spin", None) is None


def test_atom_copy():
    atom = Atom("A", energy=1.0)

    copy = atom.copy()
    assert copy.name == atom.name
    assert copy.energy == atom.energy
    atom.energy = 2.0
    assert copy.energy != atom.energy


def test_atom_to_dict():
    atom = Atom("A", energy=1.0)
    expected = {"name": "A", "color": None, "radius": 0.2, "energy": 1.0}
    actual = atom.dict()
    actual.pop("index")
    assert actual == expected


def test_atom_param_length():
    atom = Atom("A", energy=1.0)
    assert len(atom) == 3


def test_atom_iter():
    atom = Atom("A", energy=1.0)
    assert list(atom) == ["color", "radius", "energy"]


def test_atoms_equal():
    atom1 = Atom("A", energy=1.0)
    atom2 = Atom("B")
    atom3 = Atom("B", energy=1.0)
    assert atom1 != atom2
    assert atom2 == atom3
    assert atom1 == "A"
    assert atom1 != "B"
    assert atom2 == "B"
    assert atom3 == "B"



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Objects for representing the shape of a finite lattice."""

import numpy as np
import scipy.spatial
import matplotlib.pyplot as plt
import itertools
from abc import ABC, abstractmethod
from .plotting import draw_lines, draw_surfaces


class AbstractShape(ABC):
    """Abstract shape object."""

    def __init__(self, dim, pos=None):
        self.dim = dim
        self.pos = np.zeros(dim) if pos is None else np.array(pos)

    @abstractmethod
    def limits(self):
        """Returns the limits of the shape."""
        pass

    @abstractmethod
    def contains(self, points, tol=0.0):
        """Checks if the given points are contained in the shape."""
        pass

    @abstractmethod
    def plot(self, ax, color="k", lw=0.0, alpha=0.2, **kwargs):
        """Plots the contour of the shape."""
        pass

    def __repr__(self):
        return self.__class__.__name__


# noinspection PyUnresolvedReferences,PyShadowingNames
class Shape(AbstractShape):
    """General shape object.

    Examples
    --------

    Cartesian coordinates

    >>> points = np.random.uniform(-0.5, 2.5, size=(500, 2))
    >>> s = lp.Shape((2, 2))
    >>> s.limits()
    [[0.  2. ]
     [0.  2.]]

    >>> import matplotlib.pyplot as plt
    >>> mask = s.contains(points)
    >>> s.plot(plt.gca())
    >>> plt.scatter(*points[mask].T, s=3, color="g")
    >>> plt.scatter(*points[~mask].T, s=3, color="r")
    >>> plt.gca().set_aspect("equal")
    >>> plt.show()

    Angled coordinate system

    >>> s = lp.Shape((2, 2), basis=[[1, 0.2], [0, 1]])
    >>> s.limits()
    [[0.  2. ]
     [0.  2.4]]

    >>> mask = s.contains(points)
    >>> s.plot(plt.gca())
    >>> plt.scatter(*points[mask].T, s=3, color="g")
    >>> plt.scatter(*points[~mask].T, s=3, color="r")
    >>> plt.gca().set_aspect("equal")
    >>> plt.show()
    """

    def __init__(self, shape, pos=None, basis=None):
        if not hasattr(shape, "__len__"):
            shape = [shape]
        super().__init__(len(shape), pos)
        self.size = np.array(shape)
        self.basis = None if basis is None else np.array(basis)

    def _build(self):
        corners = list(itertools.product(*zip(np.zeros(self.dim), self.size)))
        corners = self.pos + np.array(corners)
        edges = None
        surfs = None
        if self.dim == 2:
            edges = np.array([[0, 1], [0, 2], [2, 3], [3, 1]])
            surfs = np.array([[0, 1, 3, 2, 0]])
        elif self.dim == 3:
            # Edge indices
            edges = np.array(
                [
                    [0, 2],
                    [2, 3],
                    [3, 1],
                    [1, 0],
                    [4, 6],
                    [6, 7],
                    [7, 5],
                    [5, 4],
                    [0, 4],
                    [2, 6],
                    [3, 7],
                    [1, 5],
                ]
            )
            # Surface indices
            surfs = np.array(
                [
                    [0, 2, 3, 1],
                    [4, 6, 7, 5],
                    [0, 4, 6, 2],
                    [2, 6, 7, 3],
                    [3, 7, 5, 1],
                    [1, 5, 4, 0],
                ]
            )
        if self.basis is not None:
            corners = np.inner(corners, self.basis.T)
        return corners, edges, surfs

    def limits(self):
        corners, _, _ = self._build()
        lims = np.array([np.min(corners, axis=0), np.max(corners, axis=0)])
        return lims.T

    def contains(self, points, tol=0.0):
        if self.basis is not None:
            points = np.inner(points, np.linalg.inv(self.basis.T))
        mask = np.logical_and(
            self.pos - tol <= points, points <= self.pos + self.size + tol
        )
        return np.all(mask, axis=1)

    def plot(self, ax, color="k", lw=0.0, alpha=0.2, **kwargs):  # pragma: no cover
        corners, edges, surfs = self._build()
        segments = corners[edges]
        lines = draw_lines(ax, segments, color=color, lw=lw)
        segments = corners[surfs]
        if self.dim < 3:
            surfaces = ax.fill(*segments.T, color=color, alpha=alpha)
        elif self.dim == 3:
            surfaces = draw_surfaces(ax, segments, color=color, alpha=alpha)
        else:
            raise NotImplementedError("Can't plot shape in D>3!")
        return lines, surfaces


# noinspection PyUnresolvedReferences,PyShadowingNames
class Circle(AbstractShape):
    """Circle shape.

    Examples
    --------

    >>> s = lp.Circle((0, 0), radius=2)
    >>> s.limits()
    [[-2.  2.]
     [-2.  2.]]

    >>> import matplotlib.pyplot as plt
    >>> points = np.random.uniform(-2, 2, size=(500, 2))
    >>> mask = s.contains(points)
    >>> s.plot(plt.gca())
    >>> plt.scatter(*points[mask].T, s=3, color="g")
    >>> plt.scatter(*points[~mask].T, s=3, color="r")
    >>> plt.gca().set_aspect("equal")
    >>> plt.show()

    """

    def __init__(self, pos, radius):
        super().__init__(len(pos), pos)
        self.radius = radius

    def limits(self):
        rad = np.full(self.dim, self.radius)
        lims = self.pos + np.array([-rad, +rad])
        return lims.T

    def contains(self, points, tol=0.0):
        dists = np.sqrt(np.sum(np.square(points - self.pos), axis=1))
        return dists <= self.radius + tol

    def plot(self, ax, color="k", lw=0.0, alpha=0.2, **kwargs):  # pragma: no cover
        xy = tuple(self.pos)
        line = plt.Circle(xy, self.radius, lw=lw, color=color, fill=False)
        ax.add_artist(line)
        surf = plt.Circle(xy, self.radius, lw=0, color=color, alpha=alpha, fill=True)
        ax.add_artist(surf)
        return line, surf


# noinspection PyUnresolvedReferences,PyShadowingNames
class Donut(AbstractShape):
    """Circle shape with cut-out in the middle.

    Examples
    --------

    >>> s = lp.Donut((0, 0), radius_outer=2, radius_inner=1)
    >>> s.limits()
    [[-2.  2.]
     [-2.  2.]]

    >>> import matplotlib.pyplot as plt
    >>> points = np.random.uniform(-2, 2, size=(500, 2))
    >>> mask = s.contains(points)
    >>> s.plot(plt.gca())
    >>> plt.scatter(*points[mask].T, s=3, color="g")
    >>> plt.scatter(*points[~mask].T, s=3, color="r")
    >>> plt.gca().set_aspect("equal")
    >>> plt.show()

    """

    def __init__(self, pos, radius_outer, radius_inner):
        super().__init__(len(pos), pos)
        self.radii = np.array([radius_inner, radius_outer])

    def limits(self):
        rad = np.full(self.dim, self.radii[1])
        lims = self.pos + np.array([-rad, +rad])
        return lims.T

    def contains(self, points, tol=1e-10):
        dists = np.sqrt(np.sum(np.square(points - self.pos), axis=1))
        return np.logical_and(
            self.radii[0] - tol <= dists, dists <= self.radii[1] + tol
        )

    def plot(self, ax, color="k", lw=0.0, alpha=0.2, **kwargs):  # pragma: no cover
        n = 100

        theta = np.linspace(0, 2 * np.pi, n, endpoint=True)
        xs = np.outer(self.radii, np.cos(theta))
        ys = np.outer(self.radii, np.sin(theta))
        # in order to have a closed area, the circles
        # should be traversed in opposite directions
        xs[1, :] = xs[1, ::-1]
        ys[1, :] = ys[1, ::-1]

        line1 = ax.plot(xs[0], ys[0], color=color, lw=lw)[0]
        line2 = ax.plot(xs[1], ys[1], color=color, lw=lw)[0]
        surf = ax.fill(np.ravel(xs), np.ravel(ys), fc=color, alpha=alpha, ec=None)

        return [line1, line2], surf


# noinspection PyUnresolvedReferences,PyShadowingNames
class ConvexHull(AbstractShape):
    """Shape defined by convex hull of arbitrary points.

    Examples
    --------

    >>> s = lp.ConvexHull([[0, 0], [2, 0], [2, 1], [1, 2], [0, 2]])
    >>> s.limits()
    [[0.  2.]
     [0.  2.]]

    >>> import matplotlib.pyplot as plt
    >>> points = np.random.uniform(-0.5, 2.5, size=(500, 2))
    >>> mask = s.contains(points)
    >>> s.plot(plt.gca())
    >>> plt.scatter(*points[mask].T, s=3, color="g")
    >>> plt.scatter(*points[~mask].T, s=3, color="r")
    >>> plt.gca().set_aspect("equal")
    >>> plt.show()

    """

    def __init__(self, points):
        dim = len(points[0])
        super().__init__(dim)
        self.hull = scipy.spatial.ConvexHull(points)

    def limits(self):
        points = self.hull.points
        return np.array([np.min(points, axis=0), np.max(points, axis=0)]).T

    def contains(self, points, tol=1e-10):
        return np.all(
            np.add(
                np.dot(points, self.hull.equations[:, :-1].T),
                self.hull.equations[:, -1],
            )
            <= tol,  # noqa: W503
            axis=1,
        )

    def plot(self, ax, color="k", lw=0.0, alpha=0.2, **kwargs):  # pragma: no cover

        if self.dim == 2:
            segments = self.hull.points[self.hull.simplices]
            lines = draw_lines(ax, segments, color=color, lw=lw)
            # segments = self.hull.points[surf]
            segments = self.hull.points[self.hull.vertices]
            surfaces = ax.fill(*segments.T, fc=color, alpha=alpha, ec=None)

        elif self.dim == 3:

            segments = np.array(
                [self.hull.points[np.append(i, i[0])] for i in self.hull.simplices]
            )
            lines = draw_lines(ax, segments, color=color, lw=lw)

            surfaces = np.array([self.hull.points[i] for i in self.hull.simplices])
            draw_surfaces(ax, surfaces, color=color, alpha=alpha)
        else:
            raise NotImplementedError("Can't plot shape in D>3!")

        return lines, surfaces



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

import math
import numpy as np
from pytest import mark
from numpy.testing import assert_array_equal, assert_allclose
from hypothesis import given, settings, assume, strategies as st
import hypothesis.extra.numpy as hnp
from lattpy import shape


dim = st.shared(st.integers(1, 3), key="d")


@given(
    hnp.arrays(np.float64, dim, elements=st.floats(0.1, 100)),
    hnp.arrays(np.float64, dim, elements=st.floats(-100, 100)),
)
def test_shape(size, pos):
    s = shape.Shape(size, pos)

    d = len(pos)
    limits = (pos + np.array([np.zeros(d), size])).T
    assert_array_equal(s.limits(), limits)

    pts = np.random.uniform(-np.min(limits) - 10, +np.max(limits) + 10, size=(100, d))
    mask = s.contains(pts, tol=0)
    for point, res in zip(pts, mask):
        expected = True
        for i in range(d):
            expected = expected and (limits[i, 0] <= point[i] <= limits[i, 1])
        assert res == expected


@given(hnp.arrays(np.float64, dim, elements=st.floats(0.1, 100)))
def test_shape_basis(size):
    basis = np.eye(len(size))
    basis[0, 0] = 2
    s = shape.Shape(size, basis=basis)

    d = len(size)
    size[0] = size[0] * 2
    limits = (np.array([np.zeros(d), size])).T
    assert_allclose(s.limits(), limits, atol=1e-10)

    pts = np.random.uniform(-np.min(limits) - 10, +np.max(limits) + 10, size=(100, d))
    mask = s.contains(pts)
    for point, res in zip(pts, mask):
        expected = True
        for i in range(d):
            expected = expected and (limits[i, 0] <= point[i] <= limits[i, 1])
        assert res == expected


@given(hnp.arrays(np.float64, 2, elements=st.floats(-100, 100)), st.floats(1, 100))
def test_circle(pos, radius):
    s = shape.Circle(pos, radius)

    limits = np.array([pos - radius, pos + radius]).T
    assert_array_equal(s.limits(), limits)

    pts = np.random.uniform(-np.min(limits) - 10, +np.max(limits) + 10, size=(100, 2))
    mask = s.contains(pts)
    for point, res in zip(pts, mask):
        diff = point - pos
        dist = math.sqrt(diff[0] * diff[0] + diff[1] * diff[1])
        assert res == (dist <= radius)


@given(
    hnp.arrays(np.float64, 2, elements=st.floats(-100, 100)),
    st.floats(50, 100),
    st.floats(0, 49),
)
def test_donut(pos, outer, inner):
    s = shape.Donut(pos, outer, inner)

    limits = np.array([pos - outer, pos + outer]).T
    assert_array_equal(s.limits(), limits)

    pts = np.random.uniform(-np.min(limits) - 10, +np.max(limits) + 10, size=(100, 2))
    mask = s.contains(pts)
    for point, res in zip(pts, mask):
        diff = point - pos
        dist = math.sqrt(diff[0] * diff[0] + diff[1] * diff[1])
        assert res == (inner <= dist <= outer)


def test_convex_hull_2d():
    points = np.array([[0, 0], [2, 0], [2, 1], [1, 1]])
    s = shape.ConvexHull(points)

    limits = np.array([[0, 2], [0, 1]])
    assert_array_equal(s.limits(), limits)

    pts = np.random.uniform(-np.min(limits) - 1, +np.max(limits) + 1, size=(100, 2))
    mask = s.contains(pts)
    for point, res in zip(pts, mask):
        x, y = point
        if x <= 1:
            # Left half: triangle
            expected = (0 <= x <= 1) and (0 <= y <= x)
        else:
            # Right: square
            expected = (1 <= x <= 2) and (0 <= y <= 1)
        assert res == expected



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Objects for representing atoms and the unitcell of a lattice."""

import itertools
from collections import abc
from typing import Union, Any, Iterator, Dict

__all__ = ["Atom"]


class Atom(abc.MutableMapping):
    """Object representing an atom of a Bravais lattice.

    Parameters
    ----------
    name : str, optional
        The name of the atom. The default is 'A'.
    radius : float, optional
        The radius of the atom in real space.
    color : str or float or array_like, optional
        The color used to visualize the atom.
    weight : float, optional
        The weight of the atom.
    **kwargs
        Additional attributes of the atom.
    """

    _counter = itertools.count()

    __slots__ = ["_index", "_name", "_weight", "_params"]

    def __init__(
        self,
        name: str = None,
        radius: float = 0.2,
        color: str = None,
        weight: float = 1.0,
        **kwargs,
    ):
        super().__init__()
        index = next(Atom._counter)
        self._index = index
        self._name = name or "A"
        self._weight = weight
        self._params = dict(color=color, radius=radius, **kwargs)

    @property
    def id(self):
        """The id of the atom."""
        return id(self)

    @property
    def index(self) -> int:
        """Return the index of the ``Atom`` instance."""
        return self._index

    @property
    def name(self) -> str:
        """Return the name of the ``Atom`` instance."""
        return self._name

    @property
    def weight(self):
        """Return the weight or the ``Atom`` instance."""
        return self._weight

    def dict(self) -> Dict[str, Any]:
        """Returns the data of the ``Atom`` instance as a dictionary."""
        data = dict(index=self._index, name=self._name)
        data.update(self._params)
        return data

    def copy(self) -> "Atom":
        """Creates a deep copy of the ``Atom`` instance."""
        return Atom(self.name, weight=self.weight, **self._params.copy())

    def get(self, key: str, default=None) -> Any:
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def is_identical(self, other: "Atom") -> bool:
        """Checks if the other ``Atom`` is identical to this one."""
        return self._name == other.name

    def __len__(self) -> int:
        """Return the length of the ``Atom`` attributes."""
        return len(self._params)

    def __iter__(self) -> Iterator[str]:
        """Iterate over the keys of the ``Atom`` attributes."""
        return iter(self._params)

    def __getitem__(self, key: str) -> Any:
        """Make ``Atom`` attributes accessable as dictionary items."""
        return self._params[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """Make ``Atom`` attributes accessable as dictionary items."""
        self._params[key] = value

    def __delitem__(self, key: str) -> None:
        """Make ``Atom`` attributes accessable as dictionary items."""
        del self._params[key]

    def __getattribute__(self, key: str) -> Any:
        """Make ``Atom`` attributes accessable as attributes."""
        key = str(key)
        if not key.startswith("_") and key in self._params.keys():
            return self._params[key]
        else:
            return super().__getattribute__(key)

    def __setattr__(self, key: str, value: Any) -> None:
        """Make ``Atom`` attributes accessable as attributes."""
        key = str(key)
        if not key.startswith("_") and key in self._params.keys():
            self._params[key] = value
        else:
            super().__setattr__(key, value)

    def __hash__(self) -> hash:
        """Make ``Atom`` instance hashable."""
        return hash(self._name)

    def __dict__(self) -> Dict[str, Any]:  # pragma: no cover
        """Return the information of the atom as a dictionary"""
        return self.dict()

    def __copy__(self) -> "Atom":  # pragma: no cover
        """Creates a deep copy of the ``Atom`` instance."""
        return self.copy()

    def __eq__(self, other: Union["Atom", str]) -> bool:
        if isinstance(other, Atom):
            return self.is_identical(other)
        else:
            return self._name == other

    def __repr__(self) -> str:
        argstr = f"{self._name}"
        paramstr = ", ".join(f"{k}={v}" for k, v in self._params.items() if v)
        if paramstr:
            argstr += ", " + paramstr
        return f"Atom({argstr}, {self.index})"



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

from hypothesis import settings

settings.register_profile("lattpy", deadline=1000)



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

import numpy as np
from numpy.testing import assert_array_equal
from hypothesis import given, assume, strategies as st
from lattpy import simple_square, simple_chain


def _construct_lattices():
    latts = list()

    latt = simple_chain()
    latt.build(10)
    latts.append(latt)

    latt = simple_square()
    latt.build((5, 5))
    latts.append(latt)

    return latts


LATTICES = _construct_lattices()


@st.composite
def lattices(draw):
    return draw(st.sampled_from(LATTICES))


@given(lattices())
def test_dim(latt):
    assert latt.dim == latt.data.dim


@given(lattices())
def test_dim(latt):
    assert len(latt.data.positions) == latt.data.num_sites


def test_num_distances():
    chain, square = LATTICES

    assert chain.data.num_distances == 2
    assert square.data.num_distances == 2


@given(lattices())
def test_copy(latt):
    copy = latt.copy()
    copy.data.remove(0)
    assert copy.data.num_sites == latt.data.num_sites - 1


@given(st.integers(0, 100))
def test_remove(site):
    latt = simple_square()
    latt.build((10, 10))
    assume(1 < site < latt.num_sites - 1)

    # find neighbor site lower than the one to remove:
    i = 0
    for i in latt.nearest_neighbors(site):
        if i < site:
            break
    assume(i > 0)
    assume(site in list(latt.nearest_neighbors(i)))

    # Get the number of neighbors of the neighbor site of `site`.
    num_neighbors = len(latt.nearest_neighbors(i))
    # Delete the site
    latt.data.remove(site)

    # check that the number of neighbors is one less than before
    assert len(latt.nearest_neighbors(i)) == (num_neighbors - 1)


def test_sort():
    latt = simple_square()
    latt.build((5, 5))

    assert np.max(np.diff(latt.data.indices, axis=0)[:, 0]) == 1

    latt.data.sort(ax=1)
    assert np.max(np.diff(latt.data.indices, axis=0)[:, 1]) == 1


def test_append():
    latt = simple_square()
    latt.build((4, 4))
    num_sites_original = latt.num_sites
    latt2 = simple_square()
    latt2.build((5, 4), pos=(5, 0))

    latt.data.append(latt2.data)

    original_pos = latt2.positions.copy()
    pos = latt.positions[num_sites_original:]
    assert_array_equal(pos, original_pos)


def test_datamap():
    # Chain
    latt = simple_chain()
    latt.build(5)

    dmap = latt.data.map()
    # fmt: off
    res = [
        True, False, True, False, False, True, False, False,
        True, False, False, True, False, False, True, False,
    ]
    # fmt: on
    assert dmap.size == len(res)
    assert_array_equal(dmap.onsite(0), res)

    # fmt: off
    res = [
        False, True, False, True, True, False, True, True,
        False, True, True, False, True, True, False, True,
    ]
    # fmt: on
    assert_array_equal(dmap.hopping(0), res)


def test_dmap_zeros():
    # Chain
    latt = simple_chain()
    latt.build(5)
    dmap = latt.dmap()

    data = dmap.zeros()
    assert data.shape == (16,)

    data = dmap.zeros(norb=1)
    assert data.shape == (16, 1, 1)

    data = dmap.zeros(norb=2)
    assert data.shape == (16, 2, 2)


def test_dmap_csr_hamiltonian():
    # Chain
    latt = simple_chain()
    latt.build(5)

    expected = 2 * np.eye(latt.num_sites)
    expected += np.eye(latt.num_sites, k=+1) + np.eye(latt.num_sites, k=-1)

    dmap = latt.dmap()
    data = np.zeros(dmap.size)
    data[dmap.onsite()] = 2
    data[dmap.hopping()] = 1
    result = dmap.build_csr(data).toarray()

    assert_array_equal(expected, result)


def test_dmap_bsr_hamiltonian():
    # Chain
    latt = simple_chain()
    latt.build(5)
    norbs = 2

    size = norbs * latt.num_sites
    expected = 2 * np.eye(size)
    expected += np.eye(size, k=+norbs) + np.eye(size, k=-norbs)

    dmap = latt.dmap()
    data = np.zeros((dmap.size, norbs, norbs))
    data[dmap.onsite()] = 2 * np.eye(norbs)
    data[dmap.hopping()] = 1 * np.eye(norbs)
    result = dmap.build_bsr(data).toarray()

    assert_array_equal(expected, result)


def test_site_mask():
    latt = simple_square()
    latt.build((4, 4))
    data = latt.data

    expect = [0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    assert_array_equal(data.site_mask([1, 0]), expect)

    expect = [0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1]
    assert_array_equal(data.site_mask([0, 1]), expect)


def test_find_sites():
    latt = simple_square()
    latt.build((4, 4))
    data = latt.data
    mins, maxs = [1, 1], [3, 3]

    expected = [6, 7, 8, 11, 12, 13, 16, 17, 18]
    assert_array_equal(data.find_sites(mins, maxs), expected)

    expected = [0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 19, 20, 21, 22, 23, 24]
    assert_array_equal(data.find_sites(mins, maxs, invert=True), expected)


def test_find_outer_sites():
    latt = simple_square()
    latt.build((4, 4))
    data = latt.data
    offset = 1

    expected = [0, 1, 2, 3, 4, 20, 21, 22, 23, 24]
    assert_array_equal(data.find_outer_sites(0, offset), expected)

    expected = [0, 4, 5, 9, 10, 14, 15, 19, 20, 24]
    assert_array_equal(data.find_outer_sites(1, offset), expected)



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
from hypothesis import given, strategies as st
import hypothesis.extra.numpy as hnp
from lattpy.utils import SiteOccupiedError, NoAtomsError, NoConnectionsError
from lattpy import Atom
from lattpy.structure import LatticeStructure

atom = Atom()


@given(st.integers(1, 3))
def test_add_atom(dim):
    latt = LatticeStructure(np.eye(dim))

    latt.add_atom()
    assert_array_equal(latt.atom_positions[0], np.zeros(dim))

    with pytest.raises(SiteOccupiedError):
        latt.add_atom()

    with pytest.raises(ValueError):
        latt.add_atom(np.zeros(4))


def test_get_alpha():
    latt = LatticeStructure(np.eye(2))
    at1 = latt.add_atom([0.0, 0.0], atom="A")
    at2 = latt.add_atom([0.5, 0.5], atom="B")

    assert latt.get_alpha("A") == [0]
    assert latt.get_alpha(at1) == 0

    assert latt.get_alpha("B") == [1]
    assert latt.get_alpha(at2) == 1

    latt = LatticeStructure(np.eye(2))
    latt.add_atom([0.0, 0.0], atom="A")
    latt.add_atom([0.5, 0.5], atom="A")
    assert latt.get_alpha("A") == [0, 1]


def test_get_atom():
    latt = LatticeStructure(np.eye(2))
    at1 = latt.add_atom([0.0, 0.0], atom="A")
    at2 = latt.add_atom([0.5, 0.5], atom="B")

    assert latt.get_atom("A") == at1
    assert latt.get_atom(0) == at1

    assert latt.get_atom("B") == at2
    assert latt.get_atom(1) == at2


def test_add_connection():
    latt = LatticeStructure(np.eye(2))
    latt.add_atom([0.0, 0.0], atom="A")
    latt.add_atom([0.5, 0.5], atom="B")

    latt.add_connection("A", "A", 1)
    latt.add_connection("A", "B", 1)
    latt.analyze()

    # Assert neighbor atom index is right
    assert all(latt.get_neighbors(alpha=0, distidx=0)[:, -1] == 1)
    assert all(latt.get_neighbors(alpha=0, distidx=1)[:, -1] == 0)
    assert all(latt.get_neighbors(alpha=1, distidx=0)[:, -1] == 0)


def test_analyze_exceptions():
    latt = LatticeStructure(np.eye(2))
    with pytest.raises(NoAtomsError):
        latt.analyze()

    latt.add_atom()
    with pytest.raises(NoConnectionsError):
        latt.analyze()


@given(hnp.arrays(np.int64, 2, elements=st.integers(0, 10)), st.integers(0, 1))
def test_get_position(nvec, alpha):
    latt = LatticeStructure(np.eye(2))
    latt.add_atom([0.0, 0.0], atom="A")
    latt.add_atom([0.5, 0.5], atom="B")

    pos = latt.translate(nvec, latt.atom_positions[alpha])
    assert_allclose(latt.get_position(nvec, alpha), pos)


@given(hnp.arrays(np.int64, (10, 2), elements=st.integers(0, 10)), st.integers(0, 1))
def test_get_positions(nvecs, alpha):
    latt = LatticeStructure(np.eye(2))
    latt.add_atom([0.0, 0.0], atom="A")
    latt.add_atom([0.5, 0.5], atom="B")

    indices = np.array([[*nvec, alpha] for nvec in nvecs])
    results = latt.get_positions(indices)
    for res, nvec in zip(results, nvecs):
        pos = latt.translate(nvec, latt.atom_positions[alpha])
        assert_allclose(res, pos)


def test_estimate_index():
    # Square lattice
    square = LatticeStructure.square()

    expected = [2, 0]
    actual = square.estimate_index([2.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0, 2]
    actual = square.estimate_index([0.0, 2.0])
    assert_array_equal(expected, actual)

    expected = [1, 2]
    actual = square.estimate_index([1.0, 2.0])
    assert_array_equal(expected, actual)

    # Rectangular lattice
    rect = LatticeStructure.rectangular(a1=2.0, a2=1.0)

    expected = [1, 0]
    actual = rect.estimate_index([2.0, 0.0])
    assert_array_equal(expected, actual)

    expected = [0, 2]
    actual = rect.estimate_index([0.0, 2.0])
    assert_array_equal(expected, actual)

    expected = [1, 1]
    actual = rect.estimate_index([2.0, 1.0])
    assert_array_equal(expected, actual)


def test_get_neighbors():
    # chain
    latt = LatticeStructure.chain()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1, 0], [-1, 0]])
    for idx in latt.get_neighbors(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[-2, 0], [2, 0]])
    for idx in latt.get_neighbors(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))

    # square
    latt = LatticeStructure.square()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1, 0, 0], [0, -1, 0], [0, 1, 0], [-1, 0, 0]])
    for idx in latt.get_neighbors(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[1, -1, 0], [-1, -1, 0], [-1, 1, 0], [1, 1, 0]])
    for idx in latt.get_neighbors(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))


def test_get_neighbor_positions():
    # chain
    latt = LatticeStructure.chain()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1], [-1]])
    for idx in latt.get_neighbor_positions(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[-2], [2]])
    for idx in latt.get_neighbor_positions(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))

    # square
    latt = LatticeStructure.square()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1, 0], [0, -1], [0, 1], [-1, 0]])
    for idx in latt.get_neighbor_positions(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[1, -1], [-1, -1], [-1, 1], [1, 1]])
    for idx in latt.get_neighbor_positions(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))


def test_get_neighbor_vectors():
    # chain
    latt = LatticeStructure.chain()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1], [-1]])
    for idx in latt.get_neighbor_vectors(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[-2], [2]])
    for idx in latt.get_neighbor_vectors(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))

    # square
    latt = LatticeStructure.square()
    latt.add_atom()
    latt.add_connections(2)
    # Nearest neighbors
    expected = np.array([[1, 0], [0, -1], [0, 1], [-1, 0]])
    for idx in latt.get_neighbor_vectors(alpha=0, distidx=0):
        assert any((expected == idx).all(axis=1))
    # Next nearest neighbors
    expected = np.array([[1, -1], [-1, -1], [-1, 1], [1, 1]])
    for idx in latt.get_neighbor_vectors(alpha=0, distidx=1):
        assert any((expected == idx).all(axis=1))


def test_get_base_atom_dict():
    latt = LatticeStructure(np.eye(2))
    ata = latt.add_atom([0, 0], atom="A")
    atb = latt.add_atom([0.5, 0], atom="B")
    latt.add_atom([0.5, 0.5], atom="B")
    result = latt.get_base_atom_dict()

    assert len(result) == 2
    assert_array_equal(result[ata], [[0, 0]])
    assert_array_equal(result[atb], [[0.5, 0.0], [0.5, 0.5]])


def test_quick_setup():
    latt = LatticeStructure(np.eye(2), atoms={(0.0, 0.0): "A"})
    assert latt.num_base == 1
    assert latt.atoms[0] == "A"
    assert_array_equal(latt.atom_positions[0], [0.0, 0.0])

    latt = LatticeStructure(np.eye(2), atoms={(0.0, 0.0): "A"}, cons=1)
    assert_array_equal(latt.num_neighbors, [4])

    latt = LatticeStructure(np.eye(2), atoms={(0.0, 0.0): "A"}, cons={("A", "A"): 1})
    assert_array_equal(latt.num_neighbors, [4])


def test_hash():
    latt1 = LatticeStructure(np.eye(2))
    latt1.add_atom()

    latt2 = LatticeStructure(np.eye(2))
    latt2.add_atom()

    assert latt1.__hash__() != latt2.__hash__()  # Since atom indices are different



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""This module contains the main `Lattice` object."""

import pickle
import logging
import warnings
import itertools
import numpy as np
from copy import deepcopy
from scipy.sparse import csr_matrix
from scipy.spatial.distance import cdist
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.axes3d import Axes3D
from typing import Union, Optional, Tuple, Iterator, Sequence, Callable
from .utils import ArrayLike, frmt_num, NotBuiltError
from .plotting import (
    subplot,
    draw_sites,
    draw_vectors,
    draw_indices,
    connection_color_array,
)
from .spatial import KDTree, distances
from .atom import Atom
from .data import LatticeData, DataMap
from .shape import AbstractShape, Shape
from .basis import basis_t
from .structure import LatticeStructure

__all__ = ["Lattice"]

logger = logging.getLogger(__name__)


def _filter_dangling(indices, positions, neighbors, dists, min_neighbors):
    num_neighbors = np.count_nonzero(np.isfinite(dists), axis=1)
    sites = np.where(num_neighbors < min_neighbors)[0]
    if len(sites) == 0:
        return indices, positions, neighbors, dists
    elif len(sites) == indices.shape[0]:
        raise ValueError("Filtering min_neighbors would result in no sites!")

    # store current invalid index
    invalid_idx = indices.shape[0]

    # Remove data from arrays
    indices = np.delete(indices, sites, axis=0)
    positions = np.delete(positions, sites, axis=0)
    neighbors = np.delete(neighbors, sites, axis=0)
    dists = np.delete(dists, sites, axis=0)

    # Update neighbor indices and distances:
    # For each removed site below the neighbor index has to be decremented once
    mask = np.isin(neighbors, sites)
    neighbors[mask] = invalid_idx
    dists[mask] = np.inf
    for count, i in enumerate(sorted(sites)):
        neighbors[neighbors > (i - count)] -= 1

    # Update invalid indices in neighbor array since number of sites changed
    num_sites = indices.shape[0]
    neighbors[neighbors == invalid_idx] = num_sites

    return indices, positions, neighbors, dists


class Lattice(LatticeStructure):
    """Main lattice object representing a Bravais lattice model.

    Combines the ``LatticeBasis`` and the ``LatticeStructure`` class and adds
    the ability to construct finite lattice models.

    .. rubric:: Inheritance

    .. inheritance-diagram:: Lattice
       :parts: 1


    Parameters
    ----------
    basis: array_like or float or LatticeBasis
        The primitive basis vectors that define the unit cell of the lattice. If a
        ``LatticeBasis`` instance is passed it is copied and used as the new basis
        of the lattice.
    **kwargs
        Key-word arguments. Used for quickly configuring a ``Lattice`` instance.
        Allowed keywords are:

        Properties:
        atoms: Dictionary containing the atoms to add to the lattice.
        cons: Dictionary containing the connections to add to the lattice.
        shape: int or tuple defining the shape of the finite size lattice to build.
        periodic: int or list defining the periodic axes to set up.

    Examples
    --------
    Two dimensional lattice with one atom in the unit cell and nearest neighbors

    >>> import lattpy as lp
    >>> latt = lp.Lattice(np.eye(2))
    >>> latt.add_atom()
    >>> latt.add_connections(1)
    >>> _ = latt.build((5, 3))
    >>> latt
    Lattice(dim: 2, num_base: 1, num_neighbors: [4], shape: [5. 3.])

    Quick-setup of the same lattice:

    >>> import lattpy as lp
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.Lattice.square(atoms={(0.0, 0.0): "A"}, cons={("A", "A"): 1})
    >>> _ = latt.build((5, 3))
    >>> _ = latt.plot()
    >>> plt.show()
    """

    def __init__(self, basis: basis_t, **kwargs):
        super().__init__(basis, **kwargs)

        # Lattice Cache
        self.data = LatticeData()
        self.shape = None
        self.pos = None
        self.periodic_axes = list()
        self.primitive = None

        if "shape" in kwargs:
            self.build(kwargs["shape"], periodic=kwargs.get("periodic", None))

    @property
    def num_sites(self) -> int:
        """int: Number of sites in lattice data (if lattice has been built)."""
        return self.data.num_sites

    @property
    def num_cells(self) -> int:
        """int: Number of unit-cells in lattice data (if lattice has been built)."""
        return np.unique(self.data.indices[:, :-1], axis=0).shape[0]

    @property
    def indices(self):
        """np.ndarray: The lattice indices of the cached lattice data."""
        return self.data.indices

    @property
    def positions(self):
        """np.ndarray: The lattice positions of the cached lattice data."""
        return self.data.positions

    def volume(self) -> float:
        """The total volume (number of cells x cell-volume) of the built lattice.

        Returns
        -------
        vol : float
            The volume of the finite lattice structure.
        """
        return self.cell_volume * np.unique(self.data.indices[:, :-1], axis=0).shape[0]

    def alpha(self, idx: int) -> int:
        """Returns the atom component of the lattice index for a site in the lattice.

        Parameters
        ----------
        idx : int
            The super-index of a site in the cached lattice data.

        Returns
        -------
        alpha : int
            The index of the atom in the unit cell.
        """
        return self.data.indices[idx, -1]

    def atom(self, idx: int) -> Atom:
        """Returns the atom of a given site in the cached lattice data.

        Parameters
        ----------
        idx : int
            The super-index of a site in the cached lattice data.

        Returns
        -------
        atom : Atom
        """
        return self._atoms[self.data.indices[idx, -1]]

    def position(self, idx: int) -> np.ndarray:
        """Returns the position of a given site in the cached lattice data.

        Parameters
        ----------
        idx : int
            The super-index of a site in the cached lattice data.

        Returns
        -------
        pos : (D, ) np.ndarray
            The position of the lattice site.
        """
        return self.data.positions[idx]

    def limits(self) -> np.ndarray:
        """Returns the spatial limits of the lattice model.

        Returns
        -------
        limits : (2, D) np.ndarray
            An array of the limits of the lattice positions. The first axis contains
            the minimum position and the second the maximum position for all dimensions.
        """
        pos = self.positions
        return np.array([np.min(pos, axis=0), np.max(pos, axis=0)])

    def relative_position(self, fractions) -> np.ndarray:
        """Computes a relitve position in the lattice model.

        Parameters
        ----------
        fractions : float or (D, ) array_like
            The position relatice to the size of the lattice. The center of the lattice
            is returned for the fractions `[0.5, ..., 0.5]`.

        Returns
        -------
        relpos : (D, ) np.ndarray
            The relative position in the lattice model.
        """
        limits = self.limits()
        return limits[0] + np.asanyarray(fractions) * (limits[1] - limits[0])

    def center(self) -> np.ndarray:
        """Returns the spatial center of the lattice.

        Returns
        -------
        center : (D, ) np.ndarray
            The position of the spatial center of the lattice.
        """
        frac = np.ones(self.dim) * 0.5
        return self.relative_position(frac)

    def center_of_gravity(self) -> np.ndarray:
        """Computes the center of gravity of the lattice model.

        Returns
        -------
        center : (D, ) np.ndarray
            The center of gravity of the lattice model.

        Notes
        -----
        Requires that the `mass` attribute is set for each atom. If no mass is set
        the default mass of `1.0` is used.
        """
        masses = [self.atom(i).get("mass", 1.0) for i in range(self.num_sites)]
        masses = np.asarray(masses)
        center = np.sum(masses[:, None] * self.positions, axis=0) / np.sum(masses)
        return center

    def superindex_from_pos(self, pos: ArrayLike, atol: float = 1e-4) -> Optional[int]:
        """Returns the super-index of a given position.

        Parameters
        ----------
        pos : (D, ) array_like
            The position of the site in cartesian coordinates.
        atol : float, optional
            The absolute tolerance for comparing positions.

        Returns
        -------
        index : int or None
            The super-index of the site in the cached lattice data.
        """
        diff = self.data.positions - np.array(pos)[None, :]
        indices = np.where(np.all(np.abs(diff) < atol, axis=1))[0]
        if len(indices) == 0:
            return None
        return indices[0]

    def superindex_from_index(self, ind: ArrayLike) -> Optional[int]:
        """Returns the super-index of a site defined by the lattice index.

        Parameters
        ----------
        ind : (D + 1, ) array_like
            The lattice index ``(n_1, ..., n_D, alpha)`` of the site.

        Returns
        -------
        index : int or None
            The super-index of the site in the cached lattice data.
        """
        diff = self.data.indices - np.array(ind)[None, :]
        indices = np.where(np.all(np.abs(diff) < 1e-4, axis=1))[0]
        if len(indices) == 0:
            return None
        return indices[0]

    def neighbors(
        self, site: int, distidx: int = None, unique: bool = False
    ) -> np.ndarray:
        """Returns the neighours of a given site in the cached lattice data.

        Parameters
        ----------
        site : int
            The super-index of a site in the cached lattice data.
        distidx : int, optional
            Index of distance to the neighbors, default is 0 (nearest neighbors).
        unique : bool, optional
            If True, each unique pair is only returned once.

        Returns
        -------
        indices : np.ndarray of int
            The super-indices of the neighbors.
        """
        return self.data.get_neighbors(site, distidx, unique=unique)

    def nearest_neighbors(self, idx: int, unique: bool = False) -> np.ndarray:
        """Returns the nearest neighors of a given site in the cached lattice data.

        Parameters
        ----------
        idx : int
            The super-index of a site in the cached lattice data.
        unique : bool, optional
            If True, each unique pair is only return once.

        Returns
        -------
        indices : (N, ) np.ndarray of int
            The super-indices of the nearest neighbors.
        """
        return self.neighbors(idx, 0, unique)

    def iter_neighbors(
        self, site: int, unique: bool = False
    ) -> Iterator[Tuple[int, np.ndarray]]:
        """Iterates over the neighbors of all distances of a given site.

        Parameters
        ----------
        site : int
            The super-index of a site in the cached lattice data.
        unique : bool, optional
            If True, each unique pair is only return once.


        Yields
        ------
        distidx : int
            The distance index of the neighbor indices.
        neighbors : (N, ) np.ndarray
            The super-indices of the neighbors for the corresponding distance level.
        """
        return self.data.iter_neighbors(site, unique)

    def check_neighbors(self, idx0: int, idx1: int) -> Union[float, None]:
        """Checks if two sites are neighbors and returns the distance level if they are.

        Parameters
        ----------
        idx0 : int
            The first super-index of a site in the cached lattice data.
        idx1 : int
            The second super-index of a site in the cached lattice data.

        Returns
        -------
        distidx : int or None
            The distance index of the two sites if they are neighbors.
        """
        for distidx in range(self.num_distances):
            if idx1 in self.neighbors(idx0, distidx):
                return distidx
        return None

    def _update_shape(self):
        limits = self.data.get_limits()
        self.shape = limits[1] - limits[0]
        self.pos = limits[0]

    def build(
        self,
        shape: Union[float, Sequence[float], AbstractShape],
        primitive: bool = False,
        pos: Union[float, Sequence[float]] = None,
        check: bool = True,
        min_neighbors: int = None,
        num_jobs: int = -1,
        periodic: Union[bool, int, Sequence[int]] = None,
        callback: Callable = None,
        dtype: Union[int, str, np.dtype] = None,
    ):
        """Constructs the indices and neighbors of a finite size lattice.

        Parameters
        ----------
        shape : (N, ) array_like or float or AbstractShape
            Shape of finite size lattice to build.
        primitive : bool, optional
            If True the shape will be multiplied by the cell size of the model.
            The default is False.
        pos : (N, ) array_like or int, optional
            Optional position of the section to build. If ``None`` the origin is used.
        check : bool, optional
            If True the positions of the translation vectors are checked and
            filtered. The default is True. This should only be disabled if
            filtered later.
        min_neighbors : int, optional
            The minimum number of neighbors a site must have. This can be used to
            remove dangling sites at the edge of the lattice.
        num_jobs : int, optional
            Number of jobs to schedule for parallel processing of neighbors.
            If ``-1`` is given all processors are used. The default is ``-1``.
        periodic : int or array_like, optional
            Optional periodic axes to set. See ``set_periodic`` for mor details.
        callback : callable, optional
            The indices and positions are passed as arguments.
        dtype : int or str or np.dtype, optional
            Optional data-type for storing the lattice indices. Using a smaller
            bit-size may help reduce memory usage. By default, the given limits are
            checked to determine the smallest possible data-type.

        Raises
        ------
        ValueError
            Raised if the dimension of the position doesn't match the dimension of
            the lattice.
        NoConnectionsError
            Raised if no connections have been set up.
        NotAnalyzedError
            Raised if the lattice distances and base-neighbors haven't been computed.
        """
        self.data.reset()
        if not isinstance(shape, AbstractShape):
            basis = self.vectors if primitive else None
            shape = Shape(shape, pos=pos, basis=basis)
            # shape = np.atleast_1d(shape)

        self._assert_connections()
        self._assert_analyzed()

        logger.debug("Building lattice: %s at %s", shape, pos)

        # Build indices and positions
        indices, positions = self.build_indices(
            shape, primitive, pos, check, callback, dtype, return_pos=True
        )

        # Compute the neighbors and distances between the sites
        neighbors, distances_ = self.compute_neighbors(indices, positions, num_jobs)
        if min_neighbors is not None:
            data = _filter_dangling(
                indices, positions, neighbors, distances_, min_neighbors
            )
            indices, positions, neighbors, distances_ = data

        # Set data of the lattice and update shape
        self.data.set(indices, positions, neighbors, distances_)
        self.primitive = primitive
        self._update_shape()

        if periodic is not None:
            self.set_periodic(periodic)

        logger.debug(
            "Lattice shape: %s (%s)",
            self.shape,
            frmt_num(self.data.nbytes, unit="iB", div=1024),
        )
        return shape

    def _build_periodic_translation_vector(self, axes, primitive=False, indices=None):
        if indices is None:
            indices = self.indices.copy()

        axes = np.atleast_1d(axes)
        if not primitive:
            # Get lattice points limits
            indices = indices.copy()[:, :-1]  # strip alpha
            indices = np.unique(indices, axis=0)
            positions = self.transform(indices)
            limits = np.array([np.min(positions, axis=0), np.max(positions, axis=0)])
            shape = limits[1] - limits[0]
            # Get periodic point
            ppoint = np.zeros(self.dim)
            for ax in axes:
                ppoint[ax] = shape[ax] + self.cell_size[ax] * 2 / 3
            # Get periodic translation vector from point
            pnvec = self.itransform(ppoint)
            pnvec = np.round(pnvec, decimals=0)
        else:
            # Get index limits
            limits = np.array([np.min(indices, axis=0), np.max(indices, axis=0)])
            idx_size = (limits[1] - limits[0])[:-1]
            # Get periodic translation vector from limits
            pnvec = np.zeros_like(idx_size, dtype=np.int64)
            for ax in axes:
                pnvec[ax] = np.floor(idx_size[ax]) + 1
        return pnvec.astype(np.int64)

    def periodic_translation_vectors(self, axes, primitive=False):
        """Constrcuts all translation vectors for periodic boundary conditions.

        Parameters
        ----------
        axes : int or (N, ) array_like
            One or multiple axises to compute the translation vectors for.
        primitive : bool, optional
            Flag if the specified axes are in cartesian or lattice coordinates.
            If ``True`` the passed position will be multiplied with the lattice vectors.
            The default is ``False`` (cartesian coordinates).

        Returns
        -------
        nvecs : list of tuple
            The translation vectors for the periodic boundary conditions.
            The first item of each element is the axis, the second the
            corresponding translation vector.
        """
        # One axis: No combinations needed
        if isinstance(axes, int) or len(axes) == 1:
            return [(axes, self._build_periodic_translation_vector(axes, primitive))]
        # Add all combinations of the periodic axis
        items = list()
        for ax in itertools.combinations_with_replacement(axes, r=2):
            nvec = self._build_periodic_translation_vector(ax, primitive)
            items.append((ax, nvec))
            # Use +/- for every axis exept the first one to ensure all corners are hit
            if not np.all(np.array(ax) == axes[0]):
                nvec2 = np.copy(nvec)
                nvec2[1:] *= -1
                items.append((ax, nvec2))
        return items

    def _build_periodic(self, indices, positions, nvec, out_ind=None, out_pos=None):
        delta_pos = self.translate(nvec)
        delta_idx = np.append(nvec, 0)
        if out_ind is not None and out_pos is not None:
            out_ind[:] = indices + delta_idx
            out_pos[:] = positions + delta_pos
        else:
            out_ind = indices + delta_idx
            out_pos = positions + delta_pos
        return out_ind, out_pos

    def kdtree(self, positions=None, eps=0.0, boxsize=None):
        if positions is None:
            positions = self.data.positions
        k = np.sum(np.sum(self._raw_num_neighbors, axis=1)) + 1
        max_dist = np.max(self.distances) + 0.1 * np.min(self._raw_distance_matrix)
        return KDTree(positions, k, max_dist, eps=eps, boxsize=boxsize)

    def _compute_pneighbors(
        self, axis, primitive=False, indices=None, positions=None, num_jobs=-1
    ):
        if indices is None:
            indices = self.data.indices
            positions = self.data.positions

        axis = np.atleast_1d(axis)
        invald_idx = len(indices)

        # Build tree
        k = np.sum(np.sum(self._raw_num_neighbors, axis=1)) + 1
        max_dist = np.max(self.distances) + 0.1 * np.min(self._raw_distance_matrix)
        tree = KDTree(positions, k, max_dist)

        # Initialize arrays
        ind_t = np.zeros_like(indices)
        pos_t = np.zeros_like(positions)

        pidx, pdists, pnvecs, paxs = dict(), dict(), dict(), dict()
        for ax, nvec in self.periodic_translation_vectors(axis, primitive):
            # Translate positions along periodic axis
            self._build_periodic(indices, positions, nvec, ind_t, pos_t)

            # Query neighbors with translated points and filter
            neighbors, distances_ = tree.query(pos_t, num_jobs, self.DIST_DECIMALS)
            neighbors, distances_ = self._filter_neighbors(
                indices, neighbors, distances_, ind_t
            )

            # Convert to dict
            idx = np.where(np.isfinite(distances_).any(axis=1))[0]
            distances_ = distances_[idx]
            neighbors = neighbors[idx]
            for i, site in enumerate(idx):
                mask = i, neighbors[i] < invald_idx
                inds = neighbors[mask]
                dists = distances_[mask]

                # Update dict for indices `inds`
                pidx.setdefault(site, list()).extend(inds)  # noqa
                pdists.setdefault(site, list()).extend(dists)
                paxs.setdefault(site, list()).extend([ax] * len(inds))
                pnvecs.setdefault(site, list()).extend([nvec] * len(inds))
                # Update dict for neighbor indices of `inds`
                for j, d in zip(inds, dists):
                    pidx.setdefault(j, list()).append(site)  # noqa
                    pdists.setdefault(j, list()).append(d)
                    paxs.setdefault(j, list()).append(ax)
                    pnvecs.setdefault(j, list()).append(-nvec)

        # Convert values of dict to np.ndarray's
        for k in list(pidx.keys()):
            pi = list(pidx[k])  # indices of periodic neighbors
            # Check if periodic neighbors is in regular neighbors
            # This occurs for small lattices and makes no sense, since the
            # sites are already neighbors
            existing_neighbors = self.neighbors(k)
            for _i, ind in enumerate(pi[:]):
                if ind in existing_neighbors:
                    pi.remove(ind)

            if len(pi):
                # Convert to arrays
                sites, ind = np.unique(pi, return_index=True)
                pidx[k] = np.array(sites)
                pdists[k] = np.array(pdists[k])[ind]
                paxs[k] = np.array(paxs[k])[ind]
                pnvecs[k] = np.array(pnvecs[k])[ind]
            else:
                # Remove periodic neighbors
                del pidx[k]
                del pdists[k]
                del paxs[k]
                del pnvecs[k]
        return pidx, pdists, pnvecs, paxs

    def set_periodic(
        self, axis: Union[bool, int, Sequence[int]] = None, primitive: bool = None
    ):
        """Sets periodic boundary conditions along the given axis.

        Parameters
        ----------
        axis : bool or int or (N, ) array_like
            One or multiple axises to apply the periodic boundary conditions.
            If the axis is ``None`` the perodic boundary conditions will be removed.
        primitive : bool, optional
            Flag if the specified axes are in cartesian or lattice coordinates.
            If ``True`` the passed position will be multiplied with the lattice vectors.
            The default is ``False`` (cartesian coordinates).
            .. deprecated:: 0.8.0
              The `primitive` argument will be removed in lattpy 0.9.0

        Raises
        ------
        NotBuiltError
            Raised if the lattice hasn't been built yet.

        Notes
        -----
        The lattice has to be built before applying the periodic boundarie conditions.
        The lattice also has to be at least three atoms big in the specified directions.
        Uses the same coordinate system (cartesian or primtive basis vectors)
        as chosen for building the lattice.
        """
        if isinstance(axis, bool):
            if axis is True:
                axis = np.arange(self.dim)
            else:
                axis = None
        if primitive is not None:
            warnings.warn(
                "The `primitive` argument is deprecated and will be removed in "
                "lattpy 0.9.0. The value for building the lattice is reused!",
                DeprecationWarning,
            )

        primitive = self.primitive

        logger.debug("Computing periodic neighbors along axis %s", axis)
        if self.shape is None:
            raise NotBuiltError()
        if axis is None:
            self.data.remove_periodic()
            self.periodic_axes = list()
        else:
            self.data.remove_periodic()
            axis = np.atleast_1d(axis)
            pidx, pdists, pnvecs, paxs = self._compute_pneighbors(axis, primitive)
            if not pidx:
                return
            self.data.set_periodic(pidx, pdists, pnvecs, paxs)
            self.periodic_axes = axis

    def _compute_connection_neighbors(self, positions1, positions2):
        # Set neighbor query parameters
        k = np.sum(np.sum(self._raw_num_neighbors, axis=1)) + 1
        max_dist = np.max(self.distances) + 0.1 * np.min(self._raw_distance_matrix)

        # Build sub-lattice tree's
        tree1 = KDTree(positions1, k=k, max_dist=max_dist)
        tree2 = KDTree(positions2, k=k, max_dist=max_dist)

        pairs = list()
        distances_ = list()
        # offset = len(positions1)
        connections = tree1.query_ball_tree(tree2, max_dist)
        for i, conns in enumerate(connections):
            if conns:
                conns = np.asarray(conns)
                dists = cdist(np.asarray([positions1[i]]), positions2[conns])[0]
                for j, dist in zip(conns, dists):
                    pairs.append((i, j))
                    # pairs.append((j, i))
                    distances_.append(dist)
                    # distances_.append(dist)

        return np.array(pairs), np.array(distances_)

    def compute_connections(self, latt):
        """Computes the connections between the current and another lattice.

        Parameters
        ----------
        latt : Lattice
            The other lattice.

        Returns
        -------
        neighbors : (N, 2) np.ndarray
            The connecting pairs between the two lattices.
            The first index of each row is the index in the current lattice data, the
            second one is the index for the other lattice ``latt``.
        distances : (N) np.ndarray
            The corresponding distances for the connections.
        """
        positions2 = latt.data.positions
        return self._compute_connection_neighbors(self.data.positions, positions2)

    def minimum_distances(self, site, primitive=None):
        """Computes the minimum distances between one site and the other lattice sites.

        This method can be used to find the distances in a lattice with
        periodic boundary conditions.

        Parameters
        ----------
        site : int
            The super-index i of a site in the cached lattice data.
        primitive : bool, optional
            Flag if the periopdic boundarey conditions are set up along cartesian or
            primitive basis vectors. The default is ``False`` (cartesian coordinates).
            .. deprecated:: 0.8.0
                The `primitive` argument will be removed in lattpy 0.9.0

        Returns
        -------
        min_dists : (N, ) np.ndarray
            The minimum distances between the lattice site i and the other sites.

        Notes
        -----
        Uses the same coordinate system (cartesian or primtive basis vectors)
        as chosen for building the lattice.
        """
        if primitive is not None:
            warnings.warn(
                "The `primitive` argument is deprecated and will be removed in "
                "lattpy 0.9.0. The value for building the lattice is reused!",
                DeprecationWarning,
            )

        positions = self.positions
        # normal distances
        dists = [distances(positions[site], positions)]
        # periodic distances (to translated site)
        paxs = self.periodic_axes
        for axs, vec in self.periodic_translation_vectors(paxs, self.primitive):
            # Get position of translated lattice point and compute distances
            translated = self.translate(vec, positions[site])
            dists.append(distances(translated, positions))
            # reverse translate direction
            translated = self.translate(-vec, positions[site])
            dists.append(distances(translated, positions))
        # get minimum distances
        return np.min(dists, axis=0)

    def _append(
        self,
        ind,
        pos,
        neighbors,
        dists,
        ax=0,
        side=+1,
        sort_axis=None,
        sort_reverse=False,
        primitive=False,
    ):

        indices2 = np.copy(ind)
        positions2 = np.copy(pos)
        neighbors2 = np.copy(neighbors)
        distances2 = np.copy(dists)
        # Build translation vector
        indices = self.data.indices if side > 0 else indices2
        nvec = self._build_periodic_translation_vector(ax, primitive, indices)
        if side <= 0:
            nvec = -1 * nvec
        vec = self.translate(nvec)

        # Store temporary data
        positions1 = self.data.positions

        # Shift data of appended lattice
        indices2[:, :-1] += nvec
        positions2 += vec

        # Append data and compute connecting neighbors
        self.data.append(indices2, positions2, neighbors2, distances2)
        pairs, distances_ = self._compute_connection_neighbors(positions1, positions2)
        offset = len(positions1)
        for (i, j), dist in zip(pairs, distances_):
            self.data.add_neighbors(i, j + offset, dist)
            self.data.add_neighbors(j + offset, i, dist)

        if sort_axis is not None:
            self.data.sort(sort_axis, reverse=sort_reverse)

        # Update the shape of the lattice
        self._update_shape()

    # noinspection PyShadowingNames
    def append(
        self, latt, ax=0, side=+1, sort_ax=None, sort_reverse=False, primitive=None
    ):
        """Append another `Lattice`-instance along an axis.

        Parameters
        ----------
        latt : Lattice
            The other lattice to append to this instance.
        ax : int, optional
            The axis along the other lattice is appended. The default is 0 (x-axis).
        side : int, optional
            The side at which the new lattice is appended. If, for example, axis 0
            is used, the other lattice is appended on the right side if ``side=+1``
            and on the left side if ``side=-1``.
        sort_ax : int, optional
            The axis to sort the lattice indices after the other lattice has been
            added. The default is the value specified for ``ax``.
        sort_reverse : bool, optional
            If True, the lattice indices are sorted in reverse order.
        primitive : bool, optional
            Flag if the periopdic boundarey conditions are set up along cartesian or
            primitive basis vectors. The default is ``False`` (cartesian coordinates).
            .. deprecated:: 0.8.0
                The `primitive` argument will be removed in lattpy 0.9.0

        Notes
        -----
        Uses the same coordinate system (cartesian or primtive basis vectors)
        as chosen for building the lattice.

        Examples
        --------
        >>> latt = Lattice(np.eye(2))
        >>> latt.add_atom(neighbors=1)
        >>> latt.build((5, 2))
        >>> latt.shape
        [5. 2.]

        >>> latt2 = Lattice(np.eye(2))
        >>> latt2.add_atom(neighbors=1)
        >>> latt2.build((2, 2))
        >>> latt2.shape
        [2. 2.]

        >>> latt.append(latt2, ax=0)
        >>> latt.shape
        [8. 2.]
        """
        if primitive is not None:
            warnings.warn(
                "The `primitive` argument is deprecated and will be removed in "
                "lattpy 0.9.0. The value for building the lattice is reused!",
                DeprecationWarning,
            )

        ind = latt.data.indices
        pos = latt.data.positions
        neighbors = latt.data.neighbors
        dists = latt.data.distvals[latt.data.distances]
        self._append(
            ind, pos, neighbors, dists, ax, side, sort_ax, sort_reverse, self.primitive
        )

    def extend(self, size, ax=0, side=1, num_jobs=1, sort_ax=None, sort_reverse=False):
        """Extend the lattice along an axis.

        Parameters
        ----------
        size : float
            The size of which the lattice will be extended in direction of ``ax``.
        ax : int, optional
            The axis along the lattice is extended. The default is 0 (x-axis).
        side : int, optional
            The side at which the new lattice is appended. If, for example, axis 0
            is used, the lattice is extended to the right side if ``side=+1``
            and to the left side if ``side=-1``.
        num_jobs : int, optional
            Number of jobs to schedule for parallel processing of neighbors for new
            sites. If ``-1`` is given all processors are used. The default is ``-1``.
        sort_ax : int, optional
            The axis to sort the lattice indices after the lattice has been extended.
            The default is the value specified for ``ax``.
        sort_reverse : bool, optional
            If True, the lattice indices are sorted in reverse order.

        Examples
        --------
        >>> latt = Lattice(np.eye(2))
        >>> latt.add_atom(neighbors=1)
        >>> latt.build((5, 2))
        >>> latt.shape
        [5. 2.]

        >>> latt.extend(2, ax=0)
        [8. 2.]

        >>> latt.extend(2, ax=1)
        [8. 5.]
        """
        # Build indices and positions of new section
        shape = np.copy(self.shape)
        shape[ax] = size
        ind, pos = self.build_indices(shape, primitive=self.primitive, return_pos=True)
        # Compute the neighbors and distances between the sites of new section
        neighbors, dists = self.compute_neighbors(ind, pos, num_jobs)
        # Append new section
        self._append(
            ind, pos, neighbors, dists, ax, side, sort_ax, sort_reverse, self.primitive
        )

    def repeat(self, num=1, ax=0, side=1, sort_ax=None, sort_reverse=False):
        """Repeat the lattice along an axis.

        Parameters
        ----------
        num : int
            The number of times the lattice will be repeated in direction ``ax``.
        ax : int, optional
            The axis along the lattice is extended. The default is 0 (x-axis).
        side : int, optional
            The side at which the new lattice is appended. If, for example, axis 0
            is used, the lattice is extended to the right side if ``side=+1``
            and to the left side if ``side=-1``.
        sort_ax : int, optional
            The axis to sort the lattice indices after the lattice has been extended.
            The default is the value specified for ``ax``.
        sort_reverse : bool, optional
            If True, the lattice indices are sorted in reverse order.

        Examples
        --------
        >>> latt = Lattice(np.eye(2))
        >>> latt.add_atom(neighbors=1)
        >>> latt.build((5, 2))
        >>> latt.shape
        [5. 2.]

        >>> latt.repeat()
        [11.  2.]

        >>> latt.repeat(3)
        [35.  2.]

        >>> latt.repeat(ax=1)
        [35.  5.]
        """
        ind = self.data.indices
        pos = self.data.positions
        neighbors = self.data.neighbors
        dists = self.data.distvals[self.data.distances]
        prim = self.primitive
        for _ in range(num):
            self._append(
                ind, pos, neighbors, dists, ax, side, sort_ax, sort_reverse, prim
            )

    def dmap(self) -> DataMap:
        """DataMap : Returns the data-map of the lattice model."""
        return self.data.map()

    def neighbor_pairs(self, unique=False):
        """Returns all neighbor pairs with their corresponding distances in the lattice.

        Parameters
        ----------
        unique : bool, optional
            If True, only unique pairs with i < j are returned. The default is False.

        Returns
        -------
        pairs : (N, 2) np.ndarray
            An array containing all neighbor pairs of the lattice. If `unique=True`,
            the first index is always smaller than the second index in each element.
        distindices : (N, ) np.ndarray
            The corresponding distance indices of the neighbor pairs.

        Examples
        --------
        >>> latt = Lattice.chain()
        >>> latt.add_atom(neighbors=1)
        >>> latt.build(5)
        >>> idx, distidx = latt.neighbor_pairs()
        >>> idx
        array([[0, 1],
               [1, 2],
               [1, 0],
               [2, 3],
               [2, 1],
               [3, 2]], dtype=uint8)

        >>> distidx
        array([0, 0, 0, 0, 0, 0], dtype=uint8)

        >>> idx, distidx = latt.neighbor_pairs(unique=True)
        >>> idx
        array([[0, 1],
               [1, 2],
               [2, 3]], dtype=uint8)

        """
        # Build index pairs and corresponding distance array
        dtype = np.min_scalar_type(self.num_sites)
        sites = np.arange(self.num_sites, dtype=dtype)
        sites_t = np.tile(sites, (self.data.neighbors.shape[1], 1)).T
        pairs = np.reshape([sites_t, self.data.neighbors], newshape=(2, -1)).T
        distindices = self.data.distances.flatten()

        # Filter pairs with invalid indices
        mask = distindices != self.data.invalid_distidx
        pairs = pairs[mask]
        distindices = distindices[mask]

        if unique:
            # Filter for unique pairs (i < j)
            mask = pairs[:, 0] < pairs[:, 1]
            pairs = pairs[mask]
            distindices = distindices[mask]

        return pairs, distindices

    def adjacency_matrix(self):
        """Computes the adjacency matrix for the neighbor data of the lattice.

        Returns
        -------
        adj_mat : (N, N) csr_matrix
            The adjacency matrix of the lattice.

        See Also
        --------
        neighbor_pairs : Generates a list of neighbor indices.

        Examples
        --------
        >>> latt = Lattice.chain()
        >>> latt.add_atom(neighbors=1)
        >>> latt.build(5)
        >>> adj_mat = latt.adjacency_matrix()
        >>> adj_mat.toarray()
        array([[0, 1, 0, 0],
               [1, 0, 1, 0],
               [0, 1, 0, 1],
               [0, 0, 1, 0]], dtype=int8)
        """
        pairs, distindices = self.neighbor_pairs(unique=False)
        rows, cols = pairs.T  # noqa
        data = distindices + 1
        return csr_matrix((data, (rows, cols)), dtype=np.int8)

    # ==================================================================================

    def copy(self) -> "Lattice":
        """Lattice : Creates a (deep) copy of the lattice instance."""
        return deepcopy(self)

    def todict(self) -> dict:
        """Creates a dictionary containing the information of the lattice instance.

        Returns
        -------
        d : dict
            The information defining the current instance.
        """
        d = super().todict()
        d["shape"] = self.shape
        return d

    def dumps(self):  # pragma: no cover
        """Creates a string containing the information of the lattice instance.

        Returns
        -------
        s : str
            The information defining the current instance.
        """
        lines = list()
        for key, values in self.todict().items():
            head = key + ":"
            lines.append(f"{head:<15}" + "; ".join(str(x) for x in values))
        return "\n".join(lines)

    def dump(self, file: Union[str, int, bytes]) -> None:  # pragma: no cover
        """Save the data of the ``Lattice`` instance.

        Parameters
        ----------
        file : str or int or bytes
            File name to store the lattice. If ``None`` the hash of the lattice is used.
        """
        if file is None:
            file = f"{self.__hash__()}.latt"
        with open(file, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, file: Union[str, int, bytes]) -> "Lattice":  # pragma: no cover
        """Load data of a saved ``Lattice`` instance.

        Parameters
        ----------
        file : str or int or bytes
            File name to load the lattice.

        Returns
        -------
        latt : Lattice
            The lattice restored from the file content.
        """
        with open(file, "rb") as f:
            latt = pickle.load(f)
        return latt

    def __hash__(self):
        import hashlib

        sha = hashlib.md5(self.dumps().encode("utf-8"))
        return int(sha.hexdigest(), 16)

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def plot(
        self,
        lw: float = None,
        margins: Union[Sequence[float], float] = 0.1,
        legend: bool = None,
        grid: bool = False,
        pscale: float = 0.5,
        show_periodic: bool = True,
        show_indices: bool = False,
        index_offset: float = 0.1,
        con_colors: Sequence = None,
        adjustable: str = "box",
        ax: Union[plt.Axes, Axes3D] = None,
        show: bool = False,
    ) -> Union[plt.Axes, Axes3D]:  # pragma: no cover
        """Plot the cached lattice.

        Parameters
        ----------
        lw : float, optional
            Line width of the neighbor connections.
        margins : Sequence[float] or float, optional
            The margins of the plot.
        legend : bool, optional
            Flag if legend is shown
        grid : bool, optional
            If True, draw a grid in the plot.
        pscale : float, optional
            The scale for drawing periodic connections. The default is half of the
            normal length.
        show_periodic : bool, optional
            If True the periodic connections will be shown.
        show_indices : bool, optional
            If True the index of the sites will be shown.
        index_offset : float, optional
            The positional offset of the index text labels. Only used if
            `show_indices=True`.
        con_colors : Sequence[tuple], optional
            list of colors to override the defautl connection color. Each element
            has to be a tuple with the first two elements being the atom indices of
            the pair and the third element the color, for example ``[(0, 0, 'r')]``.
        adjustable : None or {'box', 'datalim'}, optional
            If not None, this defines which parameter will be adjusted to meet
            the equal aspect ratio. If 'box', change the physical dimensions of
            the Axes. If 'datalim', change the x or y data limits.
            Only applied to 2D plots.
        ax : plt.Axes or plt.Axes3D or None, optional
            Parent plot. If None, a new plot is initialized.
        show : bool, optional
            If True, show the resulting plot.
        """
        logger.debug("Plotting lattice")
        if self.dim > 3:
            raise ValueError(f"Plotting in {self.dim} dimensions is not supported!")

        hopz, atomz = range(2)

        fig, ax = subplot(self.dim, adjustable, ax=ax)
        # Draw sites
        for alpha in range(self.num_base):
            atom = self.atoms[alpha]
            col = atom.color or f"C{alpha}"
            points = self.data.get_positions(alpha)
            label = atom.name
            draw_sites(ax, points, atom.radius, color=col, label=label, zorder=atomz)
        # Draw connections
        ccolor = "k"
        pcolor = "0.5"
        positions = self.positions
        hop_colors = connection_color_array(self.num_base, ccolor, con_colors)
        per_colors = connection_color_array(self.num_base, pcolor)
        for i in range(self.num_sites):
            at1 = self.alpha(i)
            p1 = positions[i]
            for j in self.data.get_neighbors(i, periodic=False, unique=True):
                p2 = positions[j]
                at2 = self.alpha(j)
                color = hop_colors[at1][at2]
                draw_vectors(ax, p2 - p1, p1, color=color, lw=lw, zorder=hopz)
            if show_periodic:
                mask = self.data.neighbor_mask(i, periodic=True)
                idx = self.data.neighbors[i, mask]
                pnvecs = self.data.pnvecs[i, mask]
                neighbor_pos = self.data.positions[idx]
                for j, x in enumerate(neighbor_pos):
                    at2 = self.alpha(idx[j])
                    x = self.translate(-pnvecs[j], x)
                    color = per_colors[at1][at2]
                    vec = pscale * (x - p1)
                    draw_vectors(ax, vec, p1, color=color, lw=lw, zorder=hopz)
        # Add index labels
        if show_indices:
            positions = [self.position(i) for i in range(self.num_sites)]
            draw_indices(ax, positions, index_offset)
        # Configure legend
        if legend is None:
            legend = self.num_base > 1
        if legend:
            ax.legend()
        # Configure grid
        if grid and self.dim < 3:
            ax.set_axisbelow(True)
            ax.grid(b=True, which="major")
        # Adjust margin
        if isinstance(margins, float):
            margins = [margins] * self.dim
        ax.margins(*margins)

        fig.tight_layout()
        if show:
            plt.show()
        return ax

    def __repr__(self) -> str:
        shape = str(self.shape) if self.shape is not None else "None"
        return (
            f"{self.__class__.__name__}("
            f"dim: {self.dim}, "
            f"num_base: {self.num_base}, "
            f"num_neighbors: {self.num_neighbors}, "
            f"shape: {shape})"
        )



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Contains miscellaneous utility methods."""

import logging
from typing import Iterable, List, Sequence, Union, Tuple
import numpy as np

__all__ = [
    "ArrayLike",
    "logger",
    "LatticeError",
    "ConfigurationError",
    "SiteOccupiedError",
    "NoAtomsError",
    "NoConnectionsError",
    "NotAnalyzedError",
    "NotBuiltError",
    "min_dtype",
    "chain",
    "create_lookup_table",
    "frmt_num",
    "frmt_bytes",
    "frmt_time",
]

# define type for numpy `array_like` types
ArrayLike = Union[int, float, Iterable, np.ndarray]


# Configure package logger
logger = logging.getLogger("lattpy")

_CH = logging.StreamHandler()
_CH.setLevel(logging.DEBUG)

_FRMT_STR = "[%(asctime)s] %(levelname)-8s - %(name)-15s - %(message)s"
_FRMT = logging.Formatter(_FRMT_STR, datefmt="%H:%M:%S")

_CH.setFormatter(_FRMT)  # Add formatter to stream handler
logger.addHandler(_CH)  # Add stream handler to package logger

logger.setLevel(logging.WARNING)  # Set initial logging level


class LatticeError(Exception):

    pass


class ConfigurationError(LatticeError):
    @property
    def msg(self):
        return self.args[0]

    @property
    def hint(self):
        return self.args[1]

    def __str__(self):
        msg, hint = self.args
        if hint:
            msg += f" ({hint})"
        return msg


class SiteOccupiedError(ConfigurationError):
    def __init__(self, atom, pos):
        super().__init__(
            f"Can't add {atom} to lattice, position {pos} already occupied!"
        )


class NoAtomsError(ConfigurationError):
    def __init__(self):
        super().__init__(
            "lattice doesn't contain any atoms use 'add_atom' to add an 'Atom'-object"
        )


class NotAnalyzedError(ConfigurationError):
    def __init__(self):
        msg = "lattice not analyzed"
        hint = (
            "call 'analyze' after adding atoms and connections or "
            "use the 'analyze' keyword of 'add_connection'"
        )
        super().__init__(msg, hint)


class NoConnectionsError(ConfigurationError):
    def __init__(self):
        msg = "base neighbors not configured"
        hint = (
            "call 'add_connection' after adding atoms or "
            "use the 'neighbors' keyword of 'add_atom'"
        )
        super().__init__(msg, hint)


class NotBuiltError(ConfigurationError):
    def __init__(self):
        msg = "lattice has not been built"
        hint = "use the 'build' method to construct a finite size lattice model"
        super().__init__(msg, hint)


def create_lookup_table(
    array: ArrayLike, dtype: Union[str, np.dtype] = np.uint8
) -> Tuple[np.ndarray, np.ndarray]:
    """Converts the given array to an array of indices linked to the unique values.

    Parameters
    ----------
    array : array_like
    dtype : int or np.dtype, optional
            Optional data-type for storing the indices of the unique values.
            By default `np.uint8` is used, since it is assumed that the
            input-array has only a few unique values.

    Returns
    -------
    values : np.ndarray
        The unique values occuring in the input-array.
    indices : np.ndarray
        The corresponding indices in the same shape as the input-array.
    """
    values = np.sort(np.unique(array))
    indices = np.zeros_like(array, dtype=dtype)
    for i, x in enumerate(values):
        mask = array == x
        indices[mask] = i
    return values, indices


def min_dtype(
    a: Union[int, float, np.ndarray, Iterable], signed: bool = True
) -> np.dtype:
    """Returns the minimum required dtype to store the given values.

    Parameters
    ----------
    a : array_like
        One or more values for determining the dtype.
        Should contain the maximal expected values.
    signed : bool, optional
        If `True` the dtype is forced to be signed. The default is `True`.

    Returns
    -------
    dtype : dtype
        The required dtype.
    """
    if signed:
        a = -np.max(np.abs(a)) - 1
    else:
        amin, amax = np.min(a), np.max(a)
        if amin < 0:
            a = -amax - 1 if abs(amin) <= amax else amin
        else:
            a = amax
    return np.dtype(np.min_scalar_type(a))


def chain(items: Sequence, cycle: bool = False) -> List:
    """Creates a chain between items

    Parameters
    ----------
    items : Sequence
        items to join to chain
    cycle : bool, optional
        cycle to the start of the chain if True, default: False

    Returns
    -------
    chain: list
        chain of items

    Example
    -------
    >>> print(chain(["x", "y", "z"]))
    [['x', 'y'], ['y', 'z']]

    >>> print(chain(["x", "y", "z"], True))
    [['x', 'y'], ['y', 'z'], ['z', 'x']]
    """
    result = list()
    for i in range(len(items) - 1):
        result.append([items[i], items[i + 1]])
    if cycle:
        result.append([items[-1], items[0]])
    return result


def frmt_num(num: float, dec: int = 1, unit: str = "", div: float = 1000.0) -> str:
    """Returns a formatted string of a number.

    Parameters
    ----------
    num : float
        The number to format.
    dec : int, optional
        Number of decimals. The default is 1.
    unit : str, optional
        Optional unit suffix. By default no unit-strinmg is used.
    div : float, optional
        The divider used for units. The default is 1000.

    Returns
    -------
    num_str: str
    """
    for prefix in ["", "k", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < div:
            return f"{num:.{dec}f}{prefix}{unit}"
        num /= div
    return f"{num:.{dec}f}Y{unit}"  # pragma: no cover


def frmt_bytes(num: float, dec: int = 1) -> str:  # pragma: no cover
    """Returns a formatted string of the number of bytes."""
    return frmt_num(num, dec, unit="iB", div=1024)


def frmt_time(seconds: float, short: bool = False, width: int = 0):  # pragma: no cover
    """Returns a formated string for a given time in seconds.

    Parameters
    ----------
    seconds : float
        Time value to format
    short : bool, optional
        Flag if short representation should be used.
    width : int, optional
        Optional minimum length of the returned string.

    Returns
    -------
    time_str: str
    """
    string = "00:00"

    # short time string
    if short:
        if seconds > 0:
            mins, secs = divmod(seconds, 60)
            if mins > 60:
                hours, mins = divmod(mins, 60)
                string = f"{hours:02.0f}:{mins:02.0f}h"
            else:
                string = f"{mins:02.0f}:{secs:02.0f}"

    # Full time strings
    else:
        if seconds < 1e-3:
            nanos = 1e6 * seconds
            string = f"{nanos:.0f}\u03BCs"
        elif seconds < 1:
            millis = 1000 * seconds
            string = f"{millis:.1f}ms"
        elif seconds < 60:
            string = f"{seconds:.1f}s"
        else:
            mins, seconds = divmod(seconds, 60)
            if mins < 60:
                string = f"{mins:.0f}:{seconds:04.1f}min"
            else:
                hours, mins = divmod(mins, 60)
                string = f"{hours:.0f}:{mins:02.0f}:{seconds:02.0f}h"

    if width > 0:
        string = f"{string:>{width}}"
    return string



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Contains plotting tools for the lattice and other related objects."""

import itertools
import numpy as np
from collections.abc import Iterable
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
import matplotlib.style as mpl_style
from matplotlib.lines import Line2D
from matplotlib.collections import LineCollection, Collection
from mpl_toolkits.mplot3d.art3d import Line3DCollection, Line3D, Poly3DCollection
from matplotlib.artist import allow_rasterization
from matplotlib import path, transforms
from typing import List
import colorcet as cc

__all__ = [
    "subplot",
    "draw_line",
    "draw_lines",
    "hide_box",
    "draw_arrows",
    "draw_vectors",
    "draw_points",
    "draw_indices",
    "draw_unit_cell",
    "draw_surfaces",
    "interpolate_to_grid",
    "draw_sites",
    "connection_color_array",
]

# Golden ratio as standard ratio for plot-figures
GOLDEN_RATIO = (np.sqrt(5) - 1.0) / 2.0

# ======================================================================================
# Formatting / Styling
# ======================================================================================


def set_color_cycler(color_cycle=cc.glasbey_category10):
    """Sets the colors of the pyplot color cycler.

    Parameters
    ----------
    color_cycle : Sequence
        A list of the colors to use in the prop cycler.
    """
    plt.rcParams["axes.prop_cycle"] = plt.cycler("color", color_cycle)


def use_mplstyle(style, color_cycle=None):
    """Update matplotlib rcparams according to style.

    Parameters
    ----------
    style : str or dict or Path or Iterable
        The style configuration.
    color_cycle : Sequence, optional
        A list of the colors to use in the prop cycler.
    """
    mpl_style.use(style)
    if color_cycle is not None:
        plt.rcParams["axes.prop_cycle"] = plt.cycler("color", color_cycle)


def set_equal_aspect(ax=None, adjustable="box"):
    """Sets the aspect ratio of the plot to equal.

    Parameters
    ----------
    ax : Axes
        The axes of the plot. If not given the current axes is used.
    adjustable : None or {'box', 'datalim'}, optional
        If not None, this defines which parameter will be adjusted to meet
        the equal aspect ratio. If 'box', change the physical dimensions of
        the Axes. If 'datalim', change the x or y data limits.

    Notes
    -----
    Setting the aspect ratio to equal is not supported for 3D plots and will
    be ignored in that case.
    """
    if ax is None:
        ax = plt.gca()
    if ax.name == "3d":
        return
    ax.set_aspect("equal", adjustable)


def hide_box(ax, axis=False):
    """Remove the box and optionally the axis of a plot.

    Parameters
    ----------
    ax : Axes
        The axes to remove the box.
    axis : bool, optional
        If True the axis are hiden as well as the box.
    """
    if ax.name == "3d":
        return

    for side in ["top", "right"]:
        ax.spines[side].set_visible(False)
    ax.xaxis.tick_bottom()
    ax.yaxis.tick_left()
    if axis:
        for side in ["left", "bottom"]:
            ax.spines[side].set_visible(False)
        ax.xaxis.set_ticks_position("none")
        ax.yaxis.set_ticks_position("none")
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.set_xticks([])
        ax.set_yticks([])


# ======================================================================================
# General Plotting
# ======================================================================================


def subplot(dim, adjustable="box", ax=None):
    """Generates a two- or three-dimensional subplot with a equal aspect ratio

    Parameters
    ----------
    dim : int
        The dimension of the plot.
    adjustable : None or {'box', 'datalim'}, optional
        If not None, this defines which parameter will be adjusted to meet
        the equal aspect ratio. If 'box', change the physical dimensions of
        the Axes. If 'datalim', change the x or y data limits.
        Only applied to 2D plots.
    ax : Axes, optional
        Existing axes to format. If an existing axes is passed no new figure is created.

    Returns
    -------
    fig : Figure
        The figure of the subplot.
    ax : Axes
        The newly created or formatted axes of the subplot.
    """
    if dim > 3:
        raise ValueError(f"Plotting in {dim} dimensions is not supported!")
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d" if dim == 3 else None)
    else:
        fig = ax.get_figure()
    set_equal_aspect(ax, adjustable)
    return fig, ax


# noinspection PyShadowingNames
def draw_line(ax, points, **kwargs):
    """Draw a line segment between multiple points.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the line segment.
    points : (N, D) np.ndarray
        A list of points between the line is drawn.
    **kwargs
        Additional keyword arguments for drawing the line.

    Returns
    -------
    coll : Line2D or Line3D
        The created line.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> points = np.array([[1, 0], [0.7, 0.7], [0, 1], [-0.7, 0.7], [-1, 0]])
    >>> _ = plotting.draw_line(ax, points)
    >>> ax.margins(0.1, 0.1)
    >>> plt.show()

    """
    dim = len(points[0])
    if dim < 3:
        line = Line2D(*points.T, **kwargs)
    elif dim == 3:
        line = Line3D(*points.T, **kwargs)
    else:
        raise ValueError(f"Can't draw line with dimension {dim}")
    ax.add_line(line)
    return line


# noinspection PyShadowingNames
def draw_lines(ax, segments, **kwargs):
    """Draw multiple line segments between points.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the lines.
    segments : array_like of (2, D) np.ndarray
        A list of point pairs between the lines are drawn.
    **kwargs
        Additional keyword arguments for drawing the lines.

    Returns
    -------
    coll: LineCollection or Line3DCollection
        The created line collection.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> segments = np.array([
    ...     [[0, 0], [1, 0]],
    ...     [[0, 1], [1, 1]],
    ...     [[0, 2], [1, 2]]
    ... ])
    >>> _ = plotting.draw_lines(ax, segments)
    >>> ax.margins(0.1, 0.1)
    >>> plt.show()
    """
    dim = len(segments[0][0])
    if dim < 3:
        coll = LineCollection(segments, **kwargs)
    elif dim == 3:
        coll = Line3DCollection(segments, **kwargs)
    else:
        raise ValueError(f"Can't draw lines with dimension {dim}")
    ax.add_collection(coll)
    return coll


# noinspection PyShadowingNames
def draw_vectors(ax, vectors, pos=None, **kwargs):
    """Draws multiple lines from an optional starting point in the given directions.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the lines.
    vectors : (N, D) np.ndarray
        The vectors to draw.
    pos : (D, ) np.ndarray, optional
        The starting position of the vectors. The default is the origin.
    **kwargs
        Additional keyword arguments for drawing the lines.

    Returns
    -------
    coll: LineCollection or Line3DCollection
        The created line collection.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> vectors = np.array([[1, 0], [0.7, 0.7], [0, 1], [-0.7, 0.7], [-1, 0]])
    >>> _ = plotting.draw_vectors(ax, vectors, [1, 0])
    >>> ax.margins(0.1, 0.1)
    >>> plt.show()
    """
    pos = pos if pos is not None else np.zeros(len(vectors[0]))
    vectors = np.atleast_2d(vectors)
    # Fix 1D case
    if vectors.shape[1] == 1:
        vectors = np.hstack((vectors, np.zeros((vectors.shape[0], 1))))
        pos = np.array([pos[0], 0])
    # Build segments
    segments = list()
    for v in vectors:
        segments.append([pos, pos + v])
    return draw_lines(ax, segments, **kwargs)


# noinspection PyShadowingNames
def draw_arrows(ax, vectors, pos=None, **kwargs):
    """Draws multiple arrows from an optional starting point in the given directions.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the arrows.
    vectors : (N, D) np.ndarray
        The vectors to draw.
    pos : (D, ) np.ndarray, optional
        The starting position of the vectors. The default is the origin.
    **kwargs
        Additional keyword arguments for drawing the arrows.

    Returns
    -------
    coll: LineCollection or Line3DCollection
        The created line collection.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> vectors = np.array([[1, 0], [0.7, 0.7], [0, 1], [-0.7, 0.7], [-1, 0]])
    >>> _ = plotting.draw_arrows(ax, vectors)
    >>> ax.margins(0.1, 0.1)
    >>> plt.show()
    """
    vectors = np.atleast_2d(vectors)
    num_vecs, dim = vectors.shape
    if pos is None:
        pos = np.zeros((num_vecs, dim))
    else:
        pos = np.atleast_2d(pos)
        if pos.shape[0] == 1:
            pos = np.tile(pos, (num_vecs, 1))
    assert len(pos) == len(vectors)
    points = pos.T
    directions = vectors.T
    end_points = (pos + vectors).T
    # Plot invisible points for datalim
    if dim == 1:
        end_points = np.append(end_points, np.zeros_like(end_points), axis=0)
        points = np.append(points, np.zeros_like(points), axis=0)
        directions = np.append(directions, np.zeros_like(directions), axis=0)
    ax.scatter(*end_points, s=0)
    # Draw arrows as quiver plot
    if dim != 3:
        kwargs.update({"angles": "xy", "scale_units": "xy", "scale": 1})
    else:
        kwargs.update({"normalize": False})
    return ax.quiver(*points, *directions, **kwargs)


# noinspection PyShadowingNames
def draw_points(ax, points, size=10, **kwargs):
    """Draws multiple points as scatter plot.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the points.
    points : (N, D) np.ndarray
        The positions of the points to draw.
    size : float, optional
        The size of the markers of the points.
    **kwargs
        Additional keyword arguments for drawing the points.

    Returns
    -------
    scat : PathCollection
        The scatter plot item.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> points = np.array([[1, 0], [0.7, 0.7], [0, 1], [-0.7, 0.7], [-1, 0]])
    >>> _ = plotting.draw_points(ax, points)
    >>> ax.margins(0.1, 0.1)
    >>> plt.show()
    """
    points = np.atleast_2d(points)
    # Fix 1D case
    if points.shape[1] == 1:
        points = np.hstack((points, np.zeros((points.shape[0], 1))))

    scat = ax.scatter(*points.T, s=size**2, **kwargs)
    # Manualy update data-limits
    # ax.ignore_existing_data_limits = True
    datalim = scat.get_datalim(ax.transData)
    ax.update_datalim(datalim)
    return scat


# noinspection PyShadowingNames
def draw_surfaces(ax, vertices, **kwargs):
    """Draws a 3D surfaces defined by a set of vertices.

    Parameters
    ----------
    ax : Axes3D
        The axes for drawing the surface.
    vertices : array_like
        The vertices defining the surface.
    **kwargs
        Additional keyword arguments for drawing the lines.

    Returns
    -------
    surf : Poly3DCollection
        The surface object.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> vertices = [[0, 0, 0], [1, 1, 0], [0.5, 0.5, 1]]
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111, projection="3d")
    >>> _ = plotting.draw_surfaces(ax, vertices, alpha=0.5)
    >>> plt.show()
    """
    if not isinstance(vertices[0][0], Iterable):
        vertices = [list(vertices)]
    poly = Poly3DCollection(vertices, **kwargs)
    ax.add_collection3d(poly)
    return poly


# noinspection PyShadowingNames
def text(ax, strings, positions, offset=None, **kwargs):
    """Adds multiple strings to a plot.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the text.
    strings : str or sequence of str
        The text to render.
    positions : (..., D) array_like
        The positions of the texts.
    offset : float or (D, ) array_like
        The offset of the positions of the text.
    **kwargs
        Additional keyword arguments for drawing the text.

    Returns
    -------
    texts : list
        The text items.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> points = np.array([[-1, 0], [-0.7, 0.7], [0, 1], [0.7, 0.7], [1, 0]])
    >>> strings = ["A", "B", "C", "D", "E"]
    >>> _ = plotting.text(ax, strings, points)
    >>> _ = ax.set_xlim(-1.5, +1.5)
    >>> _ = ax.set_ylim(-0.5, +1.5)
    >>> plt.show()
    """
    positions = np.atleast_2d(positions)
    texts = list()
    if offset is None:
        offset = np.zeros(max(2, len(positions[0])))
    elif isinstance(offset, float):
        offset = offset * np.ones(max(2, len(positions[0])))

    for s, pos in zip(strings, positions):
        if len(pos) == 1:
            pos = [pos, 0]
        tpos = np.asarray(pos) + offset
        txt = ax.text(*tpos, s=s, **kwargs)
        texts.append(txt)
    return texts


# ======================================================================================
# Lattice plotting
# ======================================================================================

# noinspection PyAbstractClass
class CircleCollection(Collection):
    """Custom circle collection

    The default matplotlib `CircleCollection` creates circles based on their
    area in screen units. This class uses the radius in data units. It behaves
    like a much faster version of a `PatchCollection` of `Circle`.
    The implementation is similar to `EllipseCollection`.
    """

    def __init__(self, radius, **kwargs):
        super().__init__(**kwargs)
        self.radius = np.atleast_1d(radius)
        self._paths = [path.Path.unit_circle()]
        self.set_transform(transforms.IdentityTransform())
        self._transforms = np.empty((0, 3, 3))

    def _set_transforms(self):
        ax = self.axes
        self._transforms = np.zeros((self.radius.size, 3, 3))
        self._transforms[:, 0, 0] = self.radius * ax.bbox.width / ax.viewLim.width
        self._transforms[:, 1, 1] = self.radius * ax.bbox.height / ax.viewLim.height
        self._transforms[:, 2, 2] = 1

    @allow_rasterization
    def draw(self, renderer):
        self._set_transforms()
        super().draw(renderer)


# noinspection PyShadowingNames
def draw_sites(ax, points, radius=0.2, **kwargs):
    """Draws multiple circles with a scaled radius.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the points.
    points : (N, D) np.ndarray
        The positions of the points to draw.
    radius : float
        The radius of the points. Scaling is only supported for 2D plots!
    **kwargs
        Additional keyword arguments for drawing the points.

    Returns
    -------
    point_coll : CircleCollection or PathCollection
        The circle or path collection.


    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> points = np.array([[1, 0], [0.7, 0.7], [0, 1], [-0.7, 0.7], [-1, 0]])
    >>> _ = plotting.draw_sites(ax, points, radius=0.2)
    >>> _ = ax.set_xlim(-1.5, +1.5)
    >>> _ = ax.set_ylim(-0.5, +1.5)
    >>> plotting.set_equal_aspect(ax)
    >>> plt.show()
    """
    points = np.atleast_2d(points)
    # Fix 1D case
    if points.shape[1] == 1:
        points = np.hstack((points, np.zeros((points.shape[0], 1))))

    dim = points.shape[1]
    if dim < 3:
        col = CircleCollection(
            radius, offsets=points, transOffset=ax.transData, **kwargs
        )
        ax.add_collection(col)
        label = kwargs.get("label", "")
        if label:
            ax.plot(
                [],
                [],
                marker="o",
                lw=0,
                color=kwargs.get("color", None),
                label=label,
                markersize=10,
            )
        datalim = col.get_datalim(ax.transData)
        datalim.x0 -= radius
        datalim.x1 += radius
        datalim.y0 -= radius
        datalim.y1 += radius
        ax.update_datalim(datalim)
        return col
    else:
        size = radius * 50
        scat = ax.scatter(*points.T, s=size**2, **kwargs)
        # Manualy update data-limits
        # ax.ignore_existing_data_limits = True
        datalim = scat.get_datalim(ax.transData)
        ax.update_datalim(datalim)
        return scat


def connection_color_array(num_base, default="k", colors=None) -> List[List]:
    """Construct color array for the connections between all atoms in a lattice.

    Parameters
    ----------
    num_base : int
        The number of atoms in the unit cell of a lattice.
    default : str or int or float or tuple
        The default color of the connections.
    colors : Sequence[tuple], optional
        list of colors to override the defautl connection color. Each element
        has to be a tuple with the first two elements being the atom indices of
        the pair and the third element the color, for example ``[(0, 0, 'r')]``.

    Returns
    -------
    color_array : List of List
        The connection color array

    Examples
    --------
    >>> connection_color_array(2, "k", colors=[(0, 1, "r")])
    [['k', 'r'], ['r', 'k']]

    """
    alphas = range(num_base)
    hop_colors = [[default for _ in alphas] for _ in alphas]
    if colors is not None and colors:
        for a1, a2, col in colors:
            hop_colors[a1][a2] = col
            hop_colors[a2][a1] = col
    return hop_colors


# noinspection PyShadowingNames
def draw_indices(ax, positions, offset=0.05, **kwargs):
    """Draws the indices of the given positions on the plot.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the text.
    positions : (..., D) array_like
        The positions of the texts.
    offset : float or (D, ) array_like
        The offset of the positions of the texts.
    **kwargs
        Additional keyword arguments for drawing the text.

    Returns
    -------
    texts : list
        The text items.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> points = np.array([[-1, 0], [-0.7, 0.7], [0, 1], [0.7, 0.7], [1, 0]])
    >>> fig, ax = plt.subplots()
    >>> _ = plotting.draw_points(ax, points)
    >>> _ = plotting.draw_indices(ax, points)
    >>> ax.margins(0.1, 0.1)
    >>> plt.show()
    """
    strings = [str(i) for i in range(len(positions))]
    va = "bottom"
    ha = "left"
    return text(ax, strings, positions, offset, ha=ha, va=va, **kwargs)


# noinspection PyShadowingNames
def draw_unit_cell(ax, vectors, outlines=True, **kwargs):
    """Draws the basis vectors and unit cell.

    Parameters
    ----------
    ax : Axes
        The axes for drawing the text.
    vectors : float or (D, D) array_like
        The vectors defining the basis.
    outlines : bool, optional
        If True the box define dby the basis vectors (unit cell) is drawn.
    **kwargs
        Additional keyword arguments for drawing the lines.

    Returns
    -------
    lines : list
        A list of the plotted lines.

    Examples
    --------
    >>> from lattpy import plotting
    >>> import matplotlib.pyplot as plt
    >>> vectors = np.array([[1, 0], [0, 1]])
    >>> fig, ax = plt.subplots()
    >>> _ = plotting.draw_unit_cell(ax, vectors)
    >>> plt.show()
    """
    dim = len(vectors)
    color = kwargs.pop("color", "k")

    arrows = draw_arrows(ax, vectors, color=color, **kwargs)
    lines = list()
    if outlines and dim > 1:
        for v, pos in itertools.permutations(vectors, r=2):
            data = np.asarray([pos, pos + v]).T
            line = ax.plot(*data, color=color, **kwargs)[0]
            lines.append(line)
        if dim == 3:
            for vecs in itertools.permutations(vectors, r=3):
                v, pos = vecs[0], np.sum(vecs[1:], axis=0)
                data = np.asarray([pos, pos + v]).T
                line = ax.plot(*data, color=color, **kwargs)[0]
                lines.append(line)
    return arrows, lines


def interpolate_to_grid(
    positions,
    values,
    num=(100, 100),
    offset=(0.0, 0.0),
    method="linear",
    fill_value=np.nan,
):
    x, y = positions.T

    # Create regular grid
    xi = np.linspace(min(x) - offset[0], max(x) + offset[0], num[0])
    yi = np.linspace(min(y) - offset[1], max(y) + offset[1], num[1])

    # Interpolate data to grid
    xx, yy = np.meshgrid(xi, yi)
    zz = griddata((x, y), values, (xi[None, :], yi[:, None]), method, fill_value)
    return xx, yy, zz



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones

import os
import time
import tracemalloc
import numpy as np
import matplotlib.pyplot as plt
import lattpy as lp

MB = 1024 * 1024
MAX_SITES = 10_000_000
RUNS = 3
MAX_POINTS = 30

overwrite = False
latts = {
    "chain": lp.simple_chain(),
    "square": lp.simple_square(),
    "hexagonal": lp.simple_hexagonal(),
    "cubic": lp.simple_cubic(),
}


class Profiler:
    """Profiler object for measuring time, memory and other stats."""

    def __init__(self):
        self._timefunc = time.perf_counter
        self._snapshot = None
        self._t0 = 0
        self._m0 = 0
        self._thread = None

    @property
    def seconds(self) -> float:
        """Returns the time since the timer has been started in seconds."""
        return self._timefunc() - self._t0

    @property
    def snapshot(self):
        return self._snapshot

    @property
    def memory(self):
        return self.get_memory() - self._m0

    def start(self):
        """Start the profiler."""
        tracemalloc.start()
        self._m0 = self.get_memory()
        self._t0 = self._timefunc()

    def stop(self) -> None:
        """Stop the profiler."""
        self._t0 = self._timefunc()
        tracemalloc.stop()

    def get_memory(self):
        self.take_snapshot()
        stats = self.statistics("filename")
        return sum(stat.size for stat in stats)

    def take_snapshot(self):
        snapshot = tracemalloc.take_snapshot()
        snapshot = snapshot.filter_traces(
            (
                tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
                tracemalloc.Filter(False, "<unknown>"),
            )
        )
        self._snapshot = snapshot
        return snapshot

    def statistics(self, group="lineno"):
        return self._snapshot.statistics(group)


def _benchmark_build_periodic(latt: lp.Lattice, sizes, runs, **kwargs):
    profiler = Profiler()
    data = np.zeros((len(sizes), 4))
    for i, size in enumerate(sizes):
        t, mem, per = 0.0, 0.0, 0.0
        for _ in range(runs):
            profiler.start()
            latt.build(np.ones(latt.dim) * size, **kwargs)
            t += profiler.seconds
            mem += profiler.memory
            profiler.start()
            latt.set_periodic(True)
            per += profiler.seconds

        data[i, 0] = latt.num_sites
        data[i, 1] = t / runs
        data[i, 2] = mem / runs
        data[i, 3] = per / runs
    return data


def bench_build_periodic(latt, runs=RUNS):
    total_sizes = np.geomspace(10, MAX_SITES, MAX_POINTS).astype(np.int64)
    if latt.dim > 1:
        sizes = np.unique(np.power(total_sizes, 1 / latt.dim).astype(np.int64))
    else:
        sizes = total_sizes
    return _benchmark_build_periodic(latt, sizes, runs, primitive=True)


def plot_benchmark_build(data):
    fig1, ax1 = plt.subplots()
    fig2, ax2 = plt.subplots()
    fig3, ax3 = plt.subplots()

    ax1.set_xscale("log")
    ax1.set_yscale("log")
    ax2.set_xscale("log")
    ax2.set_yscale("log")
    ax3.set_xscale("log")
    ax3.set_yscale("log")

    for name, arr in data.items():
        num_sites = arr[:, 0]
        line = ax1.plot(num_sites, arr[:, 1], label=name)[0]
        ax1.plot(num_sites, arr[:, 1], ls="--", color=line.get_color())
        ax2.plot(num_sites[1:], arr[1:, 2] / MB, label=name)
        ax3.plot(num_sites, arr[:, 3], label=name)

    ax1.legend()
    ax2.legend()
    ax3.legend()

    ax1.grid()
    ax2.grid()
    ax3.grid()

    ax1.set_ylabel("Time (s)")
    ax2.set_ylabel("Memory (MB)")
    ax3.set_ylabel("Time (s)")
    ax1.set_xlabel("N")
    ax2.set_xlabel("N")
    ax3.set_xlabel("N")
    ax1.set_ylim(0.001, 100)
    ax3.set_ylim(0.001, 100)
    fig1.tight_layout()
    fig2.tight_layout()
    fig3.tight_layout()

    fig1.savefig("bench_build_time.png")
    fig2.savefig("bench_build_memory.png")
    fig3.savefig("bench_periodic_time.png")


def main():
    file = "benchmark_build_periodic.npz"
    if overwrite or not os.path.exists(file):
        data = dict()
        for name, latt in latts.items():
            print("Benchmarking build:", name)
            values = bench_build_periodic(latt)
            data[name] = values
            np.savez(file, **data)
    else:
        data = np.load(file)

    plot_benchmark_build(data)
    plt.show()


if __name__ == "__main__":
    main()



# coding: utf-8
#
# This code is part of lattpy.
#
# Copyright (c) 2022, Dylan Jones
#
# This code is licensed under the MIT License. The copyright notice in the
# LICENSE file in the root directory and this permission notice shall
# be included in all copies or substantial portions of the Software.

"""Package for modeling Bravais lattices and finite lattice structures.

Submodules
----------

.. autosummary::
   lattpy.atom
   lattpy.basis
   lattpy.data
   lattpy.disptools
   lattpy.lattice
   lattpy.shape
   lattpy.spatial
   lattpy.structure
   lattpy.utils

"""

from .utils import (
    logger,
    ArrayLike,
    LatticeError,
    ConfigurationError,
    SiteOccupiedError,
    NoConnectionsError,
    NotAnalyzedError,
    NotBuiltError,
    min_dtype,
    chain,
    frmt_num,
    frmt_bytes,
    frmt_time,
)

from .spatial import (
    distance,
    distances,
    interweave,
    vindices,
    vrange,
    cell_size,
    cell_volume,
    compute_vectors,
    VoronoiTree,
    WignerSeitzCell,
)

from .shape import AbstractShape, Shape, Circle, Donut, ConvexHull
from .data import LatticeData
from .disptools import DispersionPath
from .atom import Atom
from .basis import LatticeBasis
from .structure import LatticeStructure
from .lattice import Lattice

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0"


# =========================================================================
#                             1D Lattices
# =========================================================================


def simple_chain(a=1.0, atom=None, neighbors=1):
    """Creates a 1D lattice with one atom at the origin of the unit cell.

    Parameters
    ----------
    a : float, optional
        The lattice constant (length of the basis-vector).
    atom : str or Atom, optional
        The atom to add to the lattice. If a string is passed, a new ``Atom`` instance
        is created.
    neighbors : int, optional
        The number of neighbor-distance levels, e.g. setting to 1 means only
        nearest neighbors. The default is nearest neighbors (1).

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.simple_chain()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    latt = Lattice.chain(a)
    latt.add_atom(atom=atom)
    latt.add_connections(neighbors)
    return latt


def alternating_chain(a=1.0, atom1=None, atom2=None, x0=0.0, neighbors=1):
    """Creates a 1D lattice with two atoms in the unit cell.

    Parameters
    ----------
    a : float, optional
        The lattice constant (length of the basis-vector).
    atom1 : str or Atom, optional
        The first atom to add to the lattice. If a string is passed, a new
        ``Atom`` instance is created.
    atom2 : str or Atom, optional
        The second atom to add to the lattice. If a string is passed, a new
        ``Atom`` instance is created.
    x0 : float, optional
        The offset of the atom positions in x-direction.
    neighbors : int, optional
        The number of neighbor-distance levels, e.g. setting to 1 means only
        nearest neighbors. The default is nearest neighbors (1).

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.alternating_chain()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    latt = Lattice.chain(a)
    latt.add_atom(pos=(0.0 + x0) * a, atom=atom1)
    latt.add_atom(pos=(0.5 + x0) * a, atom=atom2)
    latt.add_connections(neighbors)
    return latt


# =========================================================================
#                             2D Lattices
# =========================================================================


def simple_square(a=1.0, atom=None, neighbors=1):
    """Creates a square lattice with one atom at the origin of the unit cell.

    Parameters
    ----------
    a : float, optional
        The lattice constant (length of the basis-vector).
    atom : str or Atom, optional
        The atom to add to the lattice. If a string is passed, a new ``Atom`` instance
        is created.
    neighbors : int, optional
        The number of neighbor-distance levels, e.g. setting to 1 means only
        nearest neighbors. The default is nearest neighbors (1).

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.simple_square()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    latt = Lattice.square(a)
    latt.add_atom(atom=atom)
    latt.add_connections(neighbors)
    return latt


def simple_rectangular(a1=1.5, a2=1.0, atom=None, neighbors=2):
    """Creates a rectangular lattice with one atom at the origin of the unit cell.

    Parameters
    ----------
    a1 : float, optional
        The lattice constant in the x-direction.
    a2 : float, optional
        The lattice constant in the y-direction.
    atom : str or Atom, optional
        The atom to add to the lattice. If a string is passed, a new ``Atom`` instance
        is created.
    neighbors : int, optional
        The number of neighbor-distance levels, e.g. setting to 1 means only
        nearest neighbors. The default is nearest neighbors (2).

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.simple_rectangular()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    latt = Lattice.rectangular(a1, a2)
    latt.add_atom(atom=atom)
    latt.add_connections(neighbors)
    return latt


def simple_hexagonal(a=1.0, atom=None, neighbors=1):
    """Creates a hexagonal lattice with one atom at the origin of the unit cell.

    Parameters
    ----------
    a : float, optional
        The lattice constant (length of the basis-vector).
    atom : str or Atom, optional
        The atom to add to the lattice. If a string is passed, a new ``Atom`` instance
        is created.
    neighbors : int, optional
        The number of neighbor-distance levels, e.g. setting to 1 means only
        nearest neighbors. The default is nearest neighbors (1).

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.simple_hexagonal()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    latt = Lattice.hexagonal(a)
    latt.add_atom(atom=atom)
    latt.add_connections(neighbors)
    return latt


def honeycomb(a=1.0, atom=None):
    """Creates a honeycomb lattice with two identical atoms in the unit cell.

    Parameters
    ----------
    a : float, optional
        The lattice constant (length of the basis-vector).
    atom : str or Atom, optional
        The atom to add to the lattice. If a string is passed, a new ``Atom`` instance
        is created.

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.honeycomb()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    if not isinstance(atom, Atom):
        atom = Atom(atom, color="C0")
    latt = Lattice.hexagonal(a)
    latt.add_atom([0, 0], atom=atom)
    latt.add_atom([a, 0], atom=atom)
    latt.add_connection(0, 1, analyze=True)
    return latt


def graphene(a=1.0):
    """Creates a hexagonal lattice with two atoms in the unit cell.

    Parameters
    ----------
    a : float, optional
        The lattice constant (length of the basis-vectors).

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.graphene()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    at1 = Atom("C1")
    at2 = Atom("C2")
    latt = Lattice.hexagonal(a)
    latt.add_atom([0, 0], at1)
    latt.add_atom([a, 0], at2)
    latt.add_connection(at1, at2, analyze=True)
    return latt


# =========================================================================
#                             3D Lattices
# =========================================================================


def simple_cubic(a=1.0, atom=None, neighbors=1):
    """Creates a cubic lattice with one atom at the origin of the unit cell.

    Parameters
    ----------
    a : float, optional
        The lattice constant (length of the basis-vector).
    atom : str or Atom, optional
        The atom to add to the lattice. If a string is passed, a new ``Atom`` instance
        is created.
    neighbors : int, optional
        The number of neighbor-distance levels, e.g. setting to 1 means only
        nearest neighbors. The default is nearest neighbors (1).

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.simple_cubic()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    latt = Lattice.sc(a)
    latt.add_atom(atom=atom)
    latt.add_connections(neighbors)
    return latt


def nacl_structure(a=1.0, atom1="Na", atom2="Cl", neighbors=1):
    """Creates a NaCl lattice structure.

    Parameters
    ----------
    a : float, optional
        The lattice constant (length of the basis-vector).
    atom1 : str or Atom, optional
        The first atom to add to the lattice. If a string is passed, a new
        ``Atom`` instance is created. The default name is `Na`.
    atom2 : str or Atom, optional
        The second atom to add to the lattice. If a string is passed, a new
        ``Atom`` instance is created.. The default name is `Cl`.
    neighbors : int, optional
        The number of neighbor-distance levels, e.g. setting to 1 means only
        nearest neighbors. The default is nearest neighbors (1).

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> latt = lp.nacl_structure()
    >>> latt.plot_cell()
    >>> plt.show()

    """
    latt = Lattice.fcc(a)
    latt.add_atom(pos=[0, 0, 0], atom=atom1)
    latt.add_atom(pos=[a / 2, a / 2, a / 2], atom=atom2)
    latt.add_connections(neighbors)
    return latt


# ======================================================================================
# Other
# ======================================================================================


def finite_hypercubic(s, a=1.0, atom=None, neighbors=1, primitive=True, periodic=None):
    """Creates a d-dimensional finite lattice model with one atom in the unit cell.

    Parameters
    ----------
    s : float or Sequence[float] or AbstractShape
        The shape of the finite lattice. This also defines the dimensionality.
    a : float, optional
        The lattice constant (length of the basis-vectors).
    atom : str or Atom, optional
        The atom to add to the lattice. If a string is passed, a new ``Atom`` instance
        is created.
    neighbors : int, optional
        The number of neighbor-distance levels, e.g. setting to 1 means only
        nearest neighbors. The default is nearest neighbors (1).
    primitive : bool, optional
        If True the shape will be multiplied by the cell size of the model.
        The default is True.
    periodic : bool or int or (N, ) array_like
        One or multiple axises to apply the periodic boundary conditions.
        If the axis is ``None`` no perodic boundary conditions will be set.

    Returns
    -------
    latt : Lattice
        The configured lattice instance.

    Examples
    --------
    Simple chain:

    >>> import matplotlib.pyplot as plt
    >>> latt = lp.finite_hypercubic(4)
    >>> latt.plot()
    >>> plt.show()

    Simple square:

    >>> import matplotlib.pyplot as plt
    >>> latt = lp.finite_hypercubic((4, 2))
    >>> latt.plot()
    >>> plt.show()

    """
    import numpy as np

    if isinstance(s, (float, int)):
        s = (s,)

    dim = s.dim if isinstance(s, AbstractShape) else len(s)
    latt = Lattice(a * np.eye(dim))
    latt.add_atom(atom=atom)
    latt.add_connections(neighbors)
    latt.build(s, primitive=primitive, periodic=periodic)
    return latt
